# Copyright (c) 2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from copy import deepcopy
import datetime
from unittest import mock

from ddt import data
from ddt import ddt
import six

from cinder import exception
from cinder.objects import fields
from cinder.tests.unit import test
from cinder.tests.unit.volume.drivers.dell_emc.powermax import (
    powermax_data as tpd)
from cinder.tests.unit.volume.drivers.dell_emc.powermax import (
    powermax_fake_objects as tpfo)
from cinder.volume.drivers.dell_emc.powermax import iscsi
from cinder.volume.drivers.dell_emc.powermax import rest
from cinder.volume.drivers.dell_emc.powermax import utils
from cinder.volume import volume_types
from cinder.volume import volume_utils


@ddt
class PowerMaxUtilsTest(test.TestCase):
    def setUp(self):
        self.data = tpd.PowerMaxData()
        volume_utils.get_max_over_subscription_ratio = mock.Mock()
        super(PowerMaxUtilsTest, self).setUp()
        self.replication_device = self.data.sync_rep_device
        configuration = tpfo.FakeConfiguration(
            None, 'UtilsTests', 1, 1, san_ip='1.1.1.1', san_login='smc',
            powermax_array=self.data.array, powermax_srp='SRP_1',
            san_password='smc', san_api_port=8443,
            powermax_port_groups=[self.data.port_group_name_i],
            replication_device=self.replication_device)
        rest.PowerMaxRest._establish_rest_session = mock.Mock(
            return_value=tpfo.FakeRequestsSession())
        driver = iscsi.PowerMaxISCSIDriver(configuration=configuration)
        self.driver = driver
        self.common = self.driver.common
        self.utils = self.common.utils

    def test_get_volumetype_extra_specs(self):
        with mock.patch.object(volume_types, 'get_volume_type_extra_specs',
                               return_value={'specs'}) as type_mock:
            # path 1: volume_type_id not passed in
            self.data.test_volume.volume_type_id = (
                self.data.test_volume_type.id)
            self.utils.get_volumetype_extra_specs(self.data.test_volume)
            type_mock.assert_called_once_with(self.data.test_volume_type.id)
            type_mock.reset_mock()
            # path 2: volume_type_id passed in
            self.utils.get_volumetype_extra_specs(self.data.test_volume, '123')
            type_mock.assert_called_once_with('123')
            type_mock.reset_mock()
            # path 3: no type_id
            self.utils.get_volumetype_extra_specs(self.data.test_clone_volume)
            type_mock.assert_not_called()

    def test_get_volumetype_extra_specs_exception(self):
        extra_specs = self.utils.get_volumetype_extra_specs(
            {'name': 'no_type_id'})
        self.assertEqual({}, extra_specs)

    def test_get_host_short_name(self):
        host_under_16_chars = 'host_13_chars'
        host1 = self.utils.get_host_short_name(
            host_under_16_chars)
        self.assertEqual(host_under_16_chars, host1)

        host_over_16_chars = (
            'host_over_16_chars_host_over_16_chars_host_over_16_chars')
        # Check that the same md5 value is retrieved from multiple calls
        host2 = self.utils.get_host_short_name(
            host_over_16_chars)
        host3 = self.utils.get_host_short_name(
            host_over_16_chars)
        self.assertEqual(host2, host3)
        host_with_period = 'hostname.with.many.parts'
        ref_host_name = self.utils.generate_unique_trunc_host('hostname')
        host4 = self.utils.get_host_short_name(host_with_period)
        self.assertEqual(ref_host_name, host4)

    def test_get_volume_element_name(self):
        volume_id = 'ea95aa39-080b-4f11-9856-a03acf9112ad'
        volume_element_name = self.utils.get_volume_element_name(volume_id)
        expect_vol_element_name = ('OS-' + volume_id)
        self.assertEqual(expect_vol_element_name, volume_element_name)

    def test_truncate_string(self):
        # string is less than max number
        str_to_truncate = 'string'
        response = self.utils.truncate_string(str_to_truncate, 10)
        self.assertEqual(str_to_truncate, response)

    def test_get_default_oversubscription_ratio(self):
        default_ratio = 20.0
        max_over_sub_ratio1 = 30.0
        returned_max = self.utils.get_default_oversubscription_ratio(
            max_over_sub_ratio1)
        self.assertEqual(max_over_sub_ratio1, returned_max)
        max_over_sub_ratio2 = 0.5
        returned_max = self.utils.get_default_oversubscription_ratio(
            max_over_sub_ratio2)
        self.assertEqual(default_ratio, returned_max)

    def test_get_default_storage_group_name_slo_workload(self):
        srp_name = self.data.srp
        slo = self.data.slo
        workload = self.data.workload
        sg_name = self.utils.get_default_storage_group_name(
            srp_name, slo, workload)
        self.assertEqual(self.data.defaultstoragegroup_name, sg_name)

    def test_get_default_storage_group_name_no_slo(self):
        srp_name = self.data.srp
        slo = None
        workload = None
        sg_name = self.utils.get_default_storage_group_name(
            srp_name, slo, workload)
        self.assertEqual(self.data.default_sg_no_slo, sg_name)

    def test_get_default_storage_group_name_compr_disabled(self):
        srp_name = self.data.srp
        slo = self.data.slo
        workload = self.data.workload
        sg_name = self.utils.get_default_storage_group_name(
            srp_name, slo, workload, True)
        self.assertEqual(self.data.default_sg_compr_disabled, sg_name)

    def test_get_time_delta(self):
        start_time = 1487781721.09
        end_time = 1487781758.16
        delta = end_time - start_time
        ref_delta = six.text_type(datetime.timedelta(seconds=int(delta)))
        time_delta = self.utils.get_time_delta(start_time, end_time)
        self.assertEqual(ref_delta, time_delta)

    def test_get_short_protocol_type(self):
        # iscsi
        short_i_protocol = self.utils.get_short_protocol_type('iscsi')
        self.assertEqual('I', short_i_protocol)
        # fc
        short_f_protocol = self.utils.get_short_protocol_type('FC')
        self.assertEqual('F', short_f_protocol)
        # else
        other_protocol = self.utils.get_short_protocol_type('OTHER')
        self.assertEqual('OTHER', other_protocol)

    def test_get_temp_snap_name(self):
        source_device_id = self.data.device_id
        ref_name = self.data.temp_snapvx
        snap_name = self.utils.get_temp_snap_name(source_device_id)
        self.assertEqual(ref_name, snap_name)

    def test_get_array_and_device_id(self):
        volume = deepcopy(self.data.test_volume)
        external_ref = {u'source-name': u'00002'}
        array, device_id = self.utils.get_array_and_device_id(
            volume, external_ref)
        self.assertEqual(self.data.array, array)
        self.assertEqual('00002', device_id)
        # Test to check if device id returned is in upper case
        external_ref = {u'source-name': u'0028a'}
        __, device_id = self.utils.get_array_and_device_id(
            volume, external_ref)
        ref_device_id = u'0028A'
        self.assertEqual(ref_device_id, device_id)

    def test_get_array_and_device_id_exception(self):
        volume = deepcopy(self.data.test_volume)
        external_ref = {u'source-name': None}
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.utils.get_array_and_device_id,
                          volume, external_ref)

    @data({u'source-name': u'000001'}, {u'source-name': u'00028A'})
    def test_get_array_and_device_id_invalid_long_id(self, external_ref):
        volume = deepcopy(self.data.test_volume)
        # Test for device id more than 5 digits
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.utils.get_array_and_device_id,
                          volume, external_ref)

    @data({u'source-name': u'01'}, {u'source-name': u'028A'},
          {u'source-name': u'0001'})
    def test_get_array_and_device_id_invalid_short_id(self, external_ref):
        volume = deepcopy(self.data.test_volume)
        # Test for device id less than 5 digits
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.utils.get_array_and_device_id,
                          volume, external_ref)

    def test_get_pg_short_name(self):
        pg_under_12_chars = 'pg_11_chars'
        pg1 = self.utils.get_pg_short_name(pg_under_12_chars)
        self.assertEqual(pg_under_12_chars, pg1)

        pg_over_12_chars = 'portgroup_over_12_characters'
        # Check that the same md5 value is retrieved from multiple calls
        pg2 = self.utils.get_pg_short_name(pg_over_12_chars)
        pg3 = self.utils.get_pg_short_name(pg_over_12_chars)
        self.assertEqual(pg2, pg3)

    def test_is_compression_disabled_true(self):
        # Compression disabled in extra specs
        extra_specs = self.data.extra_specs_disable_compression
        self.assertTrue(self.utils.is_compression_disabled(extra_specs))
        # Compression disabled by no SL/WL combination
        extra_specs = deepcopy(self.data.vol_type_extra_specs_none_pool)
        self.assertTrue(self.utils.is_compression_disabled(extra_specs))
        extra_specs3 = deepcopy(extra_specs)
        extra_specs3.update({utils.DISABLECOMPRESSION: '<is> True'})
        self.assertTrue(self.utils.is_compression_disabled(extra_specs3))
        extra_specs4 = deepcopy(extra_specs)
        extra_specs4.update({utils.DISABLECOMPRESSION: 'True'})
        self.assertTrue(self.utils.is_compression_disabled(extra_specs4))

    def test_is_compression_disabled_false(self):
        # Path 1: no compression extra spec set
        extra_specs = self.data.extra_specs
        self.assertFalse(self.utils.is_compression_disabled(extra_specs))
        # Path 2: compression extra spec set to false
        extra_specs2 = deepcopy(extra_specs)
        extra_specs2.update({utils.DISABLECOMPRESSION: 'false'})
        self.assertFalse(self.utils.is_compression_disabled(extra_specs2))
        extra_specs3 = deepcopy(extra_specs)
        extra_specs3.update({utils.DISABLECOMPRESSION: '<is> False'})
        self.assertFalse(self.utils.is_compression_disabled(extra_specs3))
        extra_specs4 = deepcopy(extra_specs)
        extra_specs4.update({utils.DISABLECOMPRESSION: 'False'})
        self.assertFalse(self.utils.is_compression_disabled(extra_specs4))

    def test_change_compression_type_true(self):
        source_compr_disabled = True
        new_type_compr_disabled_1 = {
            'extra_specs': {utils.DISABLECOMPRESSION: 'false'}}
        self.assertTrue(self.utils.change_compression_type(
            source_compr_disabled, new_type_compr_disabled_1))
        new_type_compr_disabled_2 = {
            'extra_specs': {utils.DISABLECOMPRESSION: '<is> False'}}
        self.assertTrue(self.utils.change_compression_type(
            source_compr_disabled, new_type_compr_disabled_2))

    def test_change_compression_type_false(self):
        source_compr_disabled = True
        new_type_compr_disabled = {
            'extra_specs': {utils.DISABLECOMPRESSION: 'true'}}
        self.assertFalse(self.utils.change_compression_type(
            source_compr_disabled, new_type_compr_disabled))
        new_type_compr_disabled_2 = {
            'extra_specs': {utils.DISABLECOMPRESSION: '<is> True'}}
        self.assertFalse(self.utils.change_compression_type(
            source_compr_disabled, new_type_compr_disabled_2))

    def test_is_replication_enabled(self):
        is_re = self.utils.is_replication_enabled(
            self.data.vol_type_extra_specs_rep_enabled)
        self.assertTrue(is_re)
        is_re2 = self.utils.is_replication_enabled(self.data.extra_specs)
        self.assertFalse(is_re2)

    def test_get_replication_config(self):
        # Success, allow_extend false
        rep_device_list1 = [{'target_device_id': self.data.remote_array,
                             'remote_pool': self.data.srp,
                             'remote_port_group': self.data.port_group_name_f,
                             'rdf_group_label': self.data.rdf_group_name_1}]
        rep_config1 = self.utils.get_replication_config(rep_device_list1)[0]
        self.assertEqual(self.data.remote_array, rep_config1['array'])
        # Success, allow_extend true
        rep_device_list2 = rep_device_list1
        rep_device_list2[0]['allow_extend'] = 'true'
        rep_config2 = self.utils.get_replication_config(rep_device_list2)[0]
        self.assertTrue(rep_config2['allow_extend'])
        # No rep_device_list
        rep_device_list3 = []
        rep_config3 = self.utils.get_replication_config(rep_device_list3)
        self.assertIsNone(rep_config3)
        # Exception
        rep_device_list4 = [{'target_device_id': self.data.remote_array,
                             'remote_pool': self.data.srp}]
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.utils.get_replication_config, rep_device_list4)
        # Success, mode is async
        rep_device_list5 = rep_device_list2
        rep_device_list5[0]['mode'] = 'async'
        rep_config5 = self.utils.get_replication_config(rep_device_list5)[0]
        self.assertEqual(utils.REP_ASYNC, rep_config5['mode'])
        # Success, mode is metro - no other options set
        rep_device_list6 = rep_device_list5
        rep_device_list6[0]['mode'] = 'metro'
        rep_config6 = self.utils.get_replication_config(rep_device_list6)[0]
        self.assertFalse(rep_config6['metro_bias'])
        # Success, mode is metro - metro options true
        rep_device_list7 = rep_device_list6
        rep_device_list7[0].update({'metro_use_bias': 'true'})
        rep_config7 = self.utils.get_replication_config(rep_device_list7)[0]
        self.assertTrue(rep_config7['metro_bias'])
        # Success, no backend id
        self.assertIsNone(rep_config7.get(utils.BACKEND_ID))
        # Success, backend id
        rep_device_list8 = rep_device_list6
        rep_device_list8[0].update(
            {utils.BACKEND_ID: self.data.rep_backend_id_sync})
        rep_config8 = self.utils.get_replication_config(rep_device_list8)[0]
        self.assertEqual(
            self.data.rep_backend_id_sync, rep_config8[utils.BACKEND_ID])
        # Success, multi-rep
        multi_rep_device_list = self.data.multi_rep_device
        multi_rep_config = self.utils.get_replication_config(
            multi_rep_device_list)
        self.assertTrue(len(multi_rep_config) > 1)
        for rep_config in multi_rep_config:
            self.assertEqual(rep_config['array'], self.data.remote_array)

    def test_get_replication_config_sync_retries_intervals(self):
        # Default sync interval & retry values
        rep_device_list1 = [{'target_device_id': self.data.remote_array,
                             'remote_pool': self.data.srp,
                             'remote_port_group': self.data.port_group_name_f,
                             'rdf_group_label': self.data.rdf_group_name_1}]

        rep_config1 = self.utils.get_replication_config(rep_device_list1)[0]
        self.assertEqual(200, rep_config1['sync_retries'])
        self.assertEqual(3, rep_config1['sync_interval'])

        # User set interval & retry values
        rep_device_list2 = deepcopy(rep_device_list1)
        rep_device_list2[0].update({'sync_retries': 300, 'sync_interval': 1})
        rep_config2 = self.utils.get_replication_config(rep_device_list2)[0]
        self.assertEqual(300, rep_config2['sync_retries'])
        self.assertEqual(1, rep_config2['sync_interval'])

    def test_is_volume_failed_over(self):
        vol = deepcopy(self.data.test_volume)
        vol.replication_status = fields.ReplicationStatus.FAILED_OVER
        is_fo1 = self.utils.is_volume_failed_over(vol)
        self.assertTrue(is_fo1)
        is_fo2 = self.utils.is_volume_failed_over(self.data.test_volume)
        self.assertFalse(is_fo2)
        is_fo3 = self.utils.is_volume_failed_over(None)
        self.assertFalse(is_fo3)

    def test_add_legacy_pools(self):
        pools = [{'pool_name': 'Diamond+None+SRP_1+000197800111'},
                 {'pool_name': 'Diamond+OLTP+SRP_1+000197800111'}]
        new_pools = self.utils.add_legacy_pools(pools)
        ref_pools = [{'pool_name': 'Diamond+None+SRP_1+000197800111'},
                     {'pool_name': 'Diamond+OLTP+SRP_1+000197800111'},
                     {'pool_name': 'Diamond+SRP_1+000197800111'}]
        self.assertEqual(ref_pools, new_pools)

    def test_add_promotion_pools(self):
        array = self.data.array
        pools = [{'pool_name': 'Diamond+None+SRP_1+000197800111',
                  'location_info': '000197800111#SRP_1#None#Diamond'},
                 {'pool_name': 'Gold+OLTP+SRP_1+000197800111',
                  'location_info': '000197800111#SRP_1#OLTP#Gold'}]
        new_pools = self.utils.add_promotion_pools(pools, array)
        ref_pools = [{'pool_name': 'Diamond+None+SRP_1+000197800111',
                      'location_info': '000197800111#SRP_1#None#Diamond'},
                     {'pool_name': 'Gold+OLTP+SRP_1+000197800111',
                      'location_info': '000197800111#SRP_1#OLTP#Gold'},
                     {'pool_name': 'Diamond+None+SRP_1+000197800123',
                      'location_info': '000197800123#SRP_1#None#Diamond'},
                     {'pool_name': 'Gold+OLTP+SRP_1+000197800123',
                      'location_info': '000197800123#SRP_1#OLTP#Gold'}]
        self.assertEqual(ref_pools, new_pools)

    def test_update_volume_group_name(self):
        group = self.data.test_group_1
        ref_group_name = self.data.test_vol_grp_name
        vol_grp_name = self.utils.update_volume_group_name(group)
        self.assertEqual(ref_group_name, vol_grp_name)

    def test_update_volume_group_name_id_only(self):
        group = self.data.test_group_without_name
        ref_group_name = self.data.test_vol_grp_name_id_only
        vol_grp_name = self.utils.update_volume_group_name(group)
        self.assertEqual(ref_group_name, vol_grp_name)

    def test_get_volume_group_utils(self):
        array, intervals_retries = self.utils.get_volume_group_utils(
            self.data.test_group_1, interval=1, retries=1)
        ref_array = self.data.array
        self.assertEqual(ref_array, array)

    def test_update_volume_model_updates(self):
        volume_model_updates = [{'id': '1', 'status': 'available'}]
        volumes = [self.data.test_volume]
        ref_val = {'id': self.data.test_volume.id,
                   'status': 'error_deleting'}
        ret_val = self.utils.update_volume_model_updates(
            volume_model_updates, volumes, 'abc', status='error_deleting')
        self.assertEqual(ref_val, ret_val[1])

    def test_update_volume_model_updates_empty_update_list(self):
        volume_model_updates = []
        volumes = [self.data.test_volume]
        ref_val = [{'id': self.data.test_volume.id,
                    'status': 'available'}]
        ret_val = self.utils.update_volume_model_updates(
            volume_model_updates, volumes, 'abc')
        self.assertEqual(ref_val, ret_val)

    def test_update_volume_model_updates_empty_vol_list(self):
        volume_model_updates, volumes, ref_val = [], [], []
        ret_val = self.utils.update_volume_model_updates(
            volume_model_updates, volumes, 'abc')
        self.assertEqual(ref_val, ret_val)

    def test_check_replication_matched(self):
        # Check 1: Volume is not part of a group
        self.utils.check_replication_matched(
            self.data.test_volume, self.data.extra_specs)
        group_volume = deepcopy(self.data.test_volume)
        group_volume.group = self.data.test_group
        with mock.patch.object(volume_utils, 'is_group_a_type',
                               return_value=False):
            # Check 2: Both volume and group have the same rep status
            self.utils.check_replication_matched(
                group_volume, self.data.extra_specs)
            # Check 3: Volume and group have different rep status
            with mock.patch.object(self.utils, 'is_replication_enabled',
                                   return_value=True):
                self.assertRaises(exception.InvalidInput,
                                  self.utils.check_replication_matched,
                                  group_volume, self.data.extra_specs)

    def test_check_rep_status_enabled(self):
        # Check 1: not replication enabled
        with mock.patch.object(volume_utils, 'is_group_a_type',
                               return_value=False):
            self.utils.check_rep_status_enabled(self.data.test_group)
        # Check 2: replication enabled, status enabled
        with mock.patch.object(volume_utils, 'is_group_a_type',
                               return_value=True):
            self.utils.check_rep_status_enabled(self.data.test_rep_group)
            # Check 3: replication enabled, status disabled
            self.assertRaises(exception.InvalidInput,
                              self.utils.check_rep_status_enabled,
                              self.data.test_group)

    def test_get_replication_prefix(self):
        async_prefix = self.utils.get_replication_prefix(utils.REP_ASYNC)
        self.assertEqual('-RA', async_prefix)
        sync_prefix = self.utils.get_replication_prefix(utils.REP_SYNC)
        self.assertEqual('-RE', sync_prefix)
        metro_prefix = self.utils.get_replication_prefix(utils.REP_METRO)
        self.assertEqual('-RM', metro_prefix)

    def test_get_rdf_management_group_name(self):
        rep_config = {'rdf_group_label': self.data.rdf_group_name_1,
                      'mode': utils.REP_ASYNC}
        grp_name = self.utils.get_rdf_management_group_name(rep_config)
        self.assertEqual(self.data.rdf_managed_async_grp, grp_name)

    def test_is_metro_device(self):
        rep_config = {'mode': utils.REP_METRO}
        is_metro = self.utils.is_metro_device(
            rep_config, self.data.rep_extra_specs)
        self.assertTrue(is_metro)
        rep_config2 = {'mode': utils.REP_ASYNC}
        is_metro2 = self.utils.is_metro_device(
            rep_config2, self.data.rep_extra_specs)
        self.assertFalse(is_metro2)

    def test_does_vol_need_rdf_management_group(self):
        extra_specs = deepcopy(self.data.rep_extra_specs)
        extra_specs['rep_mode'] = utils.REP_SYNC
        self.assertFalse(self.utils.does_vol_need_rdf_management_group(
            extra_specs))
        extra_specs[utils.REP_MODE] = utils.REP_ASYNC
        self.assertTrue(self.utils.does_vol_need_rdf_management_group(
            extra_specs))

    def test_modify_snapshot_prefix_manage(self):
        snap_name = self.data.snapshot_id
        expected_snap_name = self.data.managed_snap_id
        updated_name = self.utils.modify_snapshot_prefix(
            snap_name, manage=True)
        self.assertEqual(expected_snap_name, updated_name)

    def test_modify_snapshot_prefix_unmanage(self):
        snap_name = self.data.managed_snap_id
        expected_snap_name = self.data.snapshot_id
        updated_name = self.utils.modify_snapshot_prefix(
            snap_name, unmanage=True)
        self.assertEqual(expected_snap_name, updated_name)

    def test_change_replication(self):
        non_rep_extra_specs = self.data.extra_specs
        rep_extra_specs = self.data.extra_specs_rep_enabled
        change_rep = self.utils.change_replication(
            non_rep_extra_specs, rep_extra_specs)
        self.assertTrue(change_rep)

    def test_change_replication_different_backend_id(self):
        rep_extra_specs_a = deepcopy(self.data.extra_specs_rep_enabled)
        rep_extra_specs_a[utils.REPLICATION_DEVICE_BACKEND_ID] = 'A'
        rep_extra_specs_b = deepcopy(self.data.extra_specs_rep_enabled)
        rep_extra_specs_b[utils.REPLICATION_DEVICE_BACKEND_ID] = 'B'
        change_rep = self.utils.change_replication(
            rep_extra_specs_a, rep_extra_specs_b)
        self.assertTrue(change_rep)

    def test_change_replication_no_change(self):
        non_rep_extra_specs_a = self.data.extra_specs
        non_rep_extra_specs_b = self.data.extra_specs
        change_rep = self.utils.change_replication(
            non_rep_extra_specs_a, non_rep_extra_specs_b)
        self.assertFalse(change_rep)

    def test_change_replication_no_change_same_backend_id(self):
        rep_extra_specs_a = deepcopy(self.data.extra_specs_rep_enabled)
        rep_extra_specs_a[utils.REPLICATION_DEVICE_BACKEND_ID] = 'A'
        rep_extra_specs_b = deepcopy(self.data.extra_specs_rep_enabled)
        rep_extra_specs_b[utils.REPLICATION_DEVICE_BACKEND_ID] = 'A'
        change_rep = self.utils.change_replication(
            rep_extra_specs_a, rep_extra_specs_b)
        self.assertFalse(change_rep)

    def test_get_child_sg_name(self):
        host_name = 'HostX'
        port_group_label = self.data.port_group_name_f
        # Slo and rep enabled
        extra_specs1 = {
            'pool_name': u'Diamond+DSS+SRP_1+000197800123',
            'slo': 'Diamond',
            'workload': 'DSS',
            'srp': 'SRP_1',
            'array': self.data.array,
            'interval': 3,
            'retries': 120,
            'replication_enabled': True,
            'rep_mode': 'Synchronous',
            utils.PORTGROUPNAME: self.data.port_group_name_f}

        child_sg_name, do_disable_compression, rep_enabled = (
            self.utils.get_child_sg_name(
                host_name, extra_specs1, port_group_label))
        re_name = self.data.storagegroup_name_f + '-RE'
        self.assertEqual(re_name, child_sg_name)
        # Disable compression
        extra_specs2 = deepcopy(self.data.extra_specs_disable_compression)
        child_sg_name, do_disable_compression, rep_enabled = (
            self.utils.get_child_sg_name(
                host_name, extra_specs2, port_group_label))
        cd_name = self.data.storagegroup_name_f + '-CD'
        self.assertEqual(cd_name, child_sg_name)
        # No slo
        extra_specs3 = deepcopy(self.data.extra_specs)
        extra_specs3[utils.SLO] = None
        child_sg_name, do_disable_compression, rep_enabled = (
            self.utils.get_child_sg_name(
                host_name, extra_specs3, port_group_label))
        self.assertEqual(self.data.no_slo_sg_name, child_sg_name)

    def test_change_multiattach(self):
        extra_specs_ma_true = {'multiattach': '<is> True'}
        extra_specs_ma_false = {'multiattach': '<is> False'}
        self.assertTrue(self.utils.change_multiattach(
            extra_specs_ma_true, extra_specs_ma_false))
        self.assertFalse(self.utils.change_multiattach(
            extra_specs_ma_true, extra_specs_ma_true))
        self.assertFalse(self.utils.change_multiattach(
            extra_specs_ma_false, extra_specs_ma_false))

    def test_is_volume_manageable(self):
        for volume in self.data.priv_vol_func_response_multi:
            self.assertTrue(
                self.utils.is_volume_manageable(volume))
        for volume in self.data.priv_vol_func_response_multi_invalid:
            self.assertFalse(
                self.utils.is_volume_manageable(volume))

    def test_is_snapshot_manageable(self):
        for volume in self.data.priv_vol_func_response_multi:
            self.assertTrue(
                self.utils.is_snapshot_manageable(volume))
        for volume in self.data.priv_vol_func_response_multi_invalid:
            self.assertFalse(
                self.utils.is_snapshot_manageable(volume))

    def test_get_volume_attached_hostname(self):

        attached_volume = deepcopy(self.data.test_volume)
        attached_volume.volume_attachment.objects = [
            self.data.test_volume_attachment]
        # Success
        hostname = self.utils.get_volume_attached_hostname(attached_volume)
        self.assertEqual('HostX', hostname)

    def test_validate_qos_input_exception(self):
        qos_extra_spec = {'total_iops_sec': 90, 'DistributionType': 'Wrong',
                          'total_bytes_sec': 100}
        input_key = 'total_iops_sec'
        sg_value = 4000
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.utils.validate_qos_input, input_key, sg_value,
                          qos_extra_spec, {})
        input_key = 'total_bytes_sec'
        sg_value = 4000
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.utils.validate_qos_input, input_key, sg_value,
                          qos_extra_spec, {})

    def test_validate_qos_distribution_type(self):
        qos_extra_spec = {'total_iops_sec': 4000, 'DistributionType': 'Always',
                          'total_bytes_sec': 4194304000}
        input_prop_dict = {'total_iops_sec': 4000}
        sg_value = 'Always'
        ret_prop_dict = self.utils.validate_qos_distribution_type(
            sg_value, qos_extra_spec, input_prop_dict)
        self.assertEqual(input_prop_dict, ret_prop_dict)

    def test_validate_qos_cast_to_int(self):
        qos_extra_spec = {'total_iops_sec': '500',
                          'total_bytes_sec': '104857600',
                          'DistributionType': 'Always'}
        property_dict = {'host_io_limit_io_sec': 500}
        input_prop_dict = {'host_io_limit_io_sec': 500,
                           'host_io_limit_mb_sec': 100}
        input_key = 'total_bytes_sec'
        ret_prop_dict = self.utils.validate_qos_input(
            input_key, None, qos_extra_spec, property_dict)
        self.assertEqual(input_prop_dict, ret_prop_dict)

    def test_validate_qos_cast_to_int_drop_fraction(self):
        qos_extra_spec = {'total_iops_sec': '500',
                          'total_bytes_sec': '105000000',
                          'DistributionType': 'Always'}
        property_dict = {'host_io_limit_io_sec': 500}
        input_prop_dict = {'host_io_limit_io_sec': 500,
                           'host_io_limit_mb_sec': 100}
        input_key = 'total_bytes_sec'
        ret_prop_dict = self.utils.validate_qos_input(
            input_key, None, qos_extra_spec, property_dict)
        self.assertEqual(input_prop_dict, ret_prop_dict)

    def test_compare_cylinders(self):
        source_cylinders = '12345'
        target_cylinders = '12345'
        self.utils.compare_cylinders(source_cylinders, target_cylinders)

    def test_compare_cylinders_target_larger(self):
        source_cylinders = '12345'
        target_cylinders = '12346'
        self.utils.compare_cylinders(source_cylinders, target_cylinders)

    def test_compare_cylinders_source_larger(self):
        source_cylinders = '12347'
        target_cylinders = '12346'
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.utils.compare_cylinders, source_cylinders,
                          target_cylinders)

    def test_get_grp_volume_model_update(self):
        volume = self.data.test_volume
        volume_dict = self.data.provider_location
        group_id = self.data.gvg_group_id
        metadata = self.data.volume_metadata

        ref_model_update_meta = {
            'id': volume.id, 'status': 'available', 'metadata': metadata,
            'provider_location': six.text_type(volume_dict)}
        act_model_update_meta = self.utils.get_grp_volume_model_update(
            volume, volume_dict, group_id, metadata)
        self.assertEqual(ref_model_update_meta, act_model_update_meta)

        ref_model_update_no_meta = {
            'id': volume.id, 'status': 'available',
            'provider_location': six.text_type(volume_dict)}
        act_model_update_no_meta = self.utils.get_grp_volume_model_update(
            volume, volume_dict, group_id)
        self.assertEqual(ref_model_update_no_meta, act_model_update_no_meta)

    def test_get_service_level_workload(self):
        # Service Level set to None
        extra_specs = deepcopy(self.data.extra_specs)
        extra_specs[utils.SLO] = None
        sl_1, wl_1 = self.utils.get_service_level_workload(extra_specs)
        self.assertEqual('None', sl_1)
        self.assertEqual('None', wl_1)
        # Service Level set to None and Workload set
        extra_specs[utils.WORKLOAD] = 'DSS'
        sl_2, wl_2 = self.utils.get_service_level_workload(extra_specs)
        self.assertEqual('None', sl_2)
        self.assertEqual('None', wl_2)
        # Service Level and Workload both set
        extra_specs[utils.SLO] = 'Diamond'
        extra_specs[utils.WORKLOAD] = 'DSS'
        sl_3, wl_3 = self.utils.get_service_level_workload(extra_specs)
        self.assertEqual('Diamond', sl_3)
        self.assertEqual('DSS', wl_3)

    def test_get_new_tags_none(self):
        list_str1 = 'finance, production,   test'
        list_str2 = 'production,test,finance'

        self.assertEqual(
            [], self.utils.get_new_tags(list_str1, list_str2))

    def test_get_new_tags_one(self):
        list_str1 = 'finance, production,   test'
        list_str2 = 'production,test'

        self.assertEqual(
            ['finance'], self.utils.get_new_tags(list_str1, list_str2))

    def test_get_new_tags_two(self):
        list_str1 = 'finance, production,   test, test2'
        list_str2 = 'production,test'

        self.assertEqual(
            ['finance', 'test2'], self.utils.get_new_tags(
                list_str1, list_str2))

    def test_get_new_tags_case(self):
        list_str1 = 'Finance, Production,   test, tEst2'
        list_str2 = 'production,test'

        self.assertEqual(
            ['Finance', 'tEst2'], self.utils.get_new_tags(
                list_str1, list_str2))

    def test_get_new_tags_empty_string_first(self):
        list_str1 = ''
        list_str2 = 'production,test'

        self.assertEqual(
            [], self.utils.get_new_tags(
                list_str1, list_str2))

    def test_get_new_tags_empty_string_second(self):
        list_str1 = 'production,test'
        list_str2 = '  '

        self.assertEqual(
            ['production', 'test'], self.utils.get_new_tags(
                list_str1, list_str2))

    def test_get_intersection(self):
        list_str1 = 'finance,production'
        list_str2 = 'production'

        common_list = self.utils._get_intersection(
            list_str1, list_str2)

        self.assertEqual(['production'], common_list)

    def test_get_intersection_unordered_list(self):
        list_str1 = 'finance,production'
        list_str2 = 'production, finance'

        common_list = (
            self.utils._get_intersection(list_str1, list_str2))

        self.assertEqual(['finance', 'production'], common_list)

    def test_verify_tag_list_good(self):
        tag_list = ['no', 'InValid', 'characters', 'dash-allowed',
                    '123', 'underscore_allowed',
                    ' leading_space', 'trailing-space ']
        self.assertTrue(self.utils.verify_tag_list(tag_list))

    def test_verify_tag_list_space(self):
        tag_list = ['bad space']
        self.assertFalse(self.utils.verify_tag_list(tag_list))

    def test_verify_tag_list_forward_slash(self):
        tag_list = ['\\forward\\slash']
        self.assertFalse(self.utils.verify_tag_list(tag_list))

    def test_verify_tag_list_square_bracket(self):
        tag_list = ['[squareBrackets]']
        self.assertFalse(self.utils.verify_tag_list(tag_list))

    def test_verify_tag_list_backward_slash(self):
        tag_list = ['/backward/slash']
        self.assertFalse(self.utils.verify_tag_list(tag_list))

    def test_verify_tag_list_curly_bracket(self):
        tag_list = ['{curlyBrackets}']
        self.assertFalse(self.utils.verify_tag_list(tag_list))

    def test_verify_tag_list_empty_list(self):
        tag_list = []
        self.assertFalse(self.utils.verify_tag_list(tag_list))

    def test_verify_tag_list_not_a_list(self):
        tag_list = '1,2,3,4'
        self.assertFalse(self.utils.verify_tag_list(tag_list))

    def test_verify_tag_list_exceeds_8(self):
        tag_list = ['1', '2', '3', '4', '5', '6', '7', '8', '9']
        self.assertFalse(self.utils.verify_tag_list(tag_list))

    def test_convert_list_to_string(self):
        input_list = ['one', 'two', 'three']
        output_string = self.utils.convert_list_to_string(input_list)
        self.assertEqual('one,two,three', output_string)

    def test_convert_list_to_string_input_string(self):
        input_list = 'one,two,three'
        output_string = self.utils.convert_list_to_string(input_list)
        self.assertEqual('one,two,three', output_string)

    def test_regex_check_case_2(self):
        test_template = 'shortHostName[:10]uuid[:5]'
        is_ok, case = self.utils.regex_check(test_template, True)
        self.assertTrue(is_ok)
        self.assertEqual('2', case)

    def test_regex_check_case_3(self):
        test_template = 'shortHostName[-10:]uuid[:5]'
        is_ok, case = self.utils.regex_check(test_template, True)
        self.assertTrue(is_ok)
        self.assertEqual('3', case)

    def test_regex_check_case_4(self):
        test_template = 'shortHostName[:7]finance'
        is_ok, case = self.utils.regex_check(test_template, True)
        self.assertTrue(is_ok)
        self.assertEqual('4', case)

    def test_regex_check_case_5(self):
        test_template = 'shortHostName[-6:]production'
        is_ok, case = self.utils.regex_check(test_template, True)
        self.assertTrue(is_ok)
        self.assertEqual('5', case)

    def test_regex_check_case_2_misspelt(self):
        test_template = 'shortHstName[:10]uuid[:5]'
        is_ok, case = self.utils.regex_check(test_template, True)
        self.assertFalse(is_ok)
        self.assertEqual('0', case)

    def test_regex_check_case_3_misspelt(self):
        test_template = 'shortHostName[-10:]uud[:5]'
        is_ok, case = self.utils.regex_check(test_template, True)
        self.assertFalse(is_ok)
        self.assertEqual('0', case)

    def test_regex_check_case_4_misspelt(self):
        test_template = 'shortHotName[:7]finance'
        is_ok, case = self.utils.regex_check(test_template, True)
        self.assertFalse(is_ok)
        self.assertEqual('0', case)

    def test_regex_check_case_5_misspelt(self):
        test_template = 'shortHstName[-6:]production'
        is_ok, case = self.utils.regex_check(test_template, True)
        self.assertFalse(is_ok)
        self.assertEqual('0', case)

    def test_regex_check_case_4_invalid_chars(self):
        test_template = 'shortHostName[:7]f*n&nce'
        is_ok, case = self.utils.regex_check(test_template, True)
        self.assertFalse(is_ok)
        self.assertEqual('0', case)

    def test_regex_check_case_5_invalid_chars(self):
        test_template = 'shortHostName[-6:]pr*ducti*n'
        is_ok, case = self.utils.regex_check(test_template, True)
        self.assertFalse(is_ok)
        self.assertEqual('0', case)

    def test_regex_check_case_2_missing_square_bracket(self):
        test_template = 'shortHostName[:10uuid[:5]'
        is_ok, case = self.utils.regex_check(test_template, True)
        self.assertFalse(is_ok)
        self.assertEqual('0', case)

    def test_regex_check_case_4_missing_square_bracket(self):
        test_template = 'shortHostName[:10finance'
        is_ok, case = self.utils.regex_check(test_template, True)
        self.assertFalse(is_ok)
        self.assertEqual('0', case)

    def test_prepare_string_entity_case_2(self):
        test_template = 'shortHostName[:10]uuid[:5]'
        altered_string = self.utils.prepare_string_entity(
            test_template, 'my_short_host_name', True)
        self.assertEqual(
            'my_short_host_name[:10]uuid[:5]',
            altered_string)

    def test_prepare_string_entity_case_3(self):
        test_template = 'shortHostName[-10:]uuid[:5]'
        altered_string = self.utils.prepare_string_entity(
            test_template, 'my_short_host_name', True)
        self.assertEqual(
            'my_short_host_name[-10:]uuid[:5]',
            altered_string)

    def test_prepare_string_entity_case_4(self):
        test_template = 'shortHostName[:7]finance'
        altered_string = self.utils.prepare_string_entity(
            test_template, 'my_short_host_name', True)
        self.assertEqual(
            'my_short_host_name[:7]finance',
            altered_string)

    def test_prepare_string_entity_case_5(self):
        test_template = 'shortHostName[-6:]production'
        altered_string = self.utils.prepare_string_entity(
            test_template, 'my_short_host_name', True)
        self.assertEqual(
            'my_short_host_name[-6:]production',
            altered_string)

    def test_prepare_string_with_uuid_case_2(self):
        test_template = 'shortHostName[:10]uuid[:5]'
        pass_two, uuid = self.utils.prepare_string_with_uuid(
            test_template, 'my_short_host_name', True)
        self.assertEqual(
            'my_short_host_name[:10]944854dce45898b544a1cb9071d3cc35[:5]',
            pass_two)
        self.assertEqual('944854dce45898b544a1cb9071d3cc35', uuid)

    def test_prepare_string_with_uuid_case_3(self):
        test_template = 'shortHostName[-10:]uuid[:5]'
        pass_two, uuid = self.utils.prepare_string_with_uuid(
            test_template, 'my_short_host_name', True)
        self.assertEqual(
            'my_short_host_name[-10:]944854dce45898b544a1cb9071d3cc35[:5]',
            pass_two)
        self.assertEqual('944854dce45898b544a1cb9071d3cc35', uuid)

    def test_check_upper_limit_short_host(self):
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.utils.check_upper_limit,
                          12, 12, True)

    def test_check_upper_limit_short_host_case_4(self):
        user_define_name = 'Little_too_long'
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.utils.check_upper_limit,
                          12, len(user_define_name), True)

    def test_validate_short_host_name_from_template_case_1(self):
        test_template = 'shortHostName'
        short_host_name = 'my_short_host'
        result_string = self.utils.validate_short_host_name_from_template(
            test_template, short_host_name)
        self.assertEqual('my_short_host', result_string)

    def test_validate_short_host_name_from_template_case_1_exceeds_16char(
            self):
        test_template = 'shortHostName'
        short_host_name = 'my_short_host_greater_than_16chars'
        result_string = self.utils.validate_short_host_name_from_template(
            test_template, short_host_name)
        self.assertEqual('6chars0bc43f914e', result_string)

    def test_validate_short_host_name_from_template_case_1_template_misspelt(
            self):
        test_template = 'shortHstName'
        short_host_name = 'my_short_host'
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.utils.validate_short_host_name_from_template,
                          test_template, short_host_name)

    def test_validate_short_host_name_from_template_case_2(self):
        test_template = 'shortHostName[:10]uuid[:5]'
        short_host_name = 'my_short_host_name'
        result_string = self.utils.validate_short_host_name_from_template(
            test_template, short_host_name)
        self.assertEqual('my_short_h94485', result_string)

    def test_validate_short_host_name_from_template_case_2_shorter_than(self):
        test_template = 'shortHostName[:10]uuid[:5]'
        short_host_name = 'HostX'
        result_string = self.utils.validate_short_host_name_from_template(
            test_template, short_host_name)
        self.assertEqual('HostX699ea', result_string)

    def test_validate_short_host_name_from_template_case_3(self):
        test_template = 'shortHostName[-10:]uuid[:5]'
        short_host_name = 'my_short_host_name'
        result_string = self.utils.validate_short_host_name_from_template(
            test_template, short_host_name)
        self.assertEqual('_host_name94485', result_string)

    def test_validate_short_host_name_from_template_case_3_shorter_than(self):
        test_template = 'shortHostName[-10:]uuid[:5]'
        short_host_name = 'HostX'
        result_string = self.utils.validate_short_host_name_from_template(
            test_template, short_host_name)
        self.assertEqual('HostX699ea', result_string)

    def test_validate_short_host_name_from_template_case_4(self):
        test_template = 'shortHostName[:7]finance'
        short_host_name = 'my_short_host_name'
        result_string = self.utils.validate_short_host_name_from_template(
            test_template, short_host_name)
        self.assertEqual('my_shorfinance', result_string)

    def test_validate_short_host_name_from_template_case_5(self):
        test_template = 'shortHostName[-6:]production'
        short_host_name = 'my_short_host_name'
        result_string = self.utils.validate_short_host_name_from_template(
            test_template, short_host_name)
        self.assertEqual('t_nameproduction', result_string)

    def test_validate_short_host_name_exception_missing_minus(self):
        test_template = 'shortHostName[6:]production'
        short_host_name = 'my_short_host_name'
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.utils.validate_short_host_name_from_template,
                          test_template, short_host_name)

    def test_validate_port_group_from_template_case_1(self):
        test_template = 'portGroupName'
        port_group_name = 'my_pg'
        result_string = self.utils.validate_port_group_name_from_template(
            test_template, port_group_name)
        self.assertEqual('my_pg', result_string)

    def test_validate_port_group_from_template_case_1_long(self):
        test_template = 'portGroupName'
        port_group_name = 'my_port_group_name'
        result_string = self.utils.validate_port_group_name_from_template(
            test_template, port_group_name)
        self.assertEqual('p_name5ba163', result_string)

    def test_validate_port_group_from_template_case_1_misspelt(self):
        test_template = 'portGr*upName'
        port_group_name = 'my_port_group_name'
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.utils.validate_port_group_name_from_template,
                          test_template, port_group_name)

    def test_validate_port_group_from_template_case_2(self):
        test_template = 'portGroupName[:6]uuid[:5]'
        port_group_name = 'my_port_group_name'
        result_string = self.utils.validate_port_group_name_from_template(
            test_template, port_group_name)
        self.assertEqual('my_por3b02c', result_string)

    def test_validate_port_group_from_template_case_3(self):
        test_template = 'portGroupName[-6:]uuid[:5]'
        port_group_name = 'my_port_group_name'
        result_string = self.utils.validate_port_group_name_from_template(
            test_template, port_group_name)
        self.assertEqual('p_name3b02c', result_string)

    def test_validate_port_group_from_template_case_4(self):
        test_template = 'portGroupName[:6]test'
        port_group_name = 'my_port_group_name'
        result_string = self.utils.validate_port_group_name_from_template(
            test_template, port_group_name)
        self.assertEqual('my_portest', result_string)

    def test_validate_port_group_from_template_case_5(self):
        test_template = 'portGroupName[-7:]test'
        port_group_name = 'my_port_group_name'
        result_string = self.utils.validate_port_group_name_from_template(
            test_template, port_group_name)
        self.assertEqual('up_nametest', result_string)

    def test_validate_port_group_name_exception_missing_minus(self):
        test_template = 'portGroupName[6:]test'
        port_group_name = 'my_port_group_name'
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.utils.validate_port_group_name_from_template,
                          test_template, port_group_name)

    def test_validate_port_group_name_exception_chars_exceeded(self):
        test_template = 'portGroupName[:10]test'
        port_group_name = 'my_port_group_name'
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.utils.validate_port_group_name_from_template,
                          test_template, port_group_name)

    def test_get_port_name_label_default(self):
        port_name_in = 'my_port_group_name'
        port_group_template = 'portGroupName'
        port_name_out = self.utils.get_port_name_label(
            port_name_in, port_group_template)
        self.assertEqual('p_name5ba163', port_name_out)

    def test_get_port_name_label_template(self):
        port_name_in = 'my_port_group_name'
        port_group_template = 'portGroupName[-6:]uuid[:5]'
        port_name_out = self.utils.get_port_name_label(
            port_name_in, port_group_template)
        self.assertEqual('p_name3b02c', port_name_out)

    def test_get_rdf_managed_storage_group(self):
        rdf_component_dict = ('OS-23_24_007-Asynchronous-rdf-sg',
                              {'prefix': 'OS',
                               'rdf_label': '23_24_007',
                               'sync_mode': 'Asynchronous',
                               'after_mode': 'rdf-sg'})

        async_rdf_details = (
            self.utils.get_rdf_managed_storage_group(
                self.data.volume_details_attached_async))
        self.assertEqual(rdf_component_dict, async_rdf_details)

    def test_get_storage_group_component_dict_no_slo(self):
        """Test for get_storage_group_component_dict.

        REST and no SLO.
        """
        sg_no_slo = 'OS-myhost-No_SLO-os-iscsi-pg'
        component_dict = self.utils.get_storage_group_component_dict(
            sg_no_slo)
        self.assertEqual('myhost', component_dict['host'])
        self.assertEqual('OS', component_dict['prefix'])
        self.assertEqual('No_SLO', component_dict['no_slo'])
        self.assertEqual('os-iscsi-pg', component_dict['portgroup'])
        self.assertIsNone(component_dict['sloworkload'])
        self.assertIsNone(component_dict['srp'])

    def test_get_storage_group_component_dict_slo_workload_2(self):
        """Test for get_storage_group_component_dict.

        SLO, workload and test 2.
        """
        sg_slo_workload = 'OS-myhost-SRP_1-DiamodOLTP-os-iscsi-pg-RE'
        component_dict = self.utils.get_storage_group_component_dict(
            sg_slo_workload)
        self.assertEqual('OS', component_dict['prefix'])
        self.assertEqual('myhost', component_dict['host'])
        self.assertEqual('SRP_1', component_dict['srp'])
        self.assertEqual('os-iscsi-pg', component_dict['portgroup'])
        self.assertEqual('DiamodOLTP', component_dict['sloworkload'])
        self.assertIsNone(component_dict['no_slo'])

    def test_get_storage_group_component_dict_compression_disabled(self):
        """Test for get_storage_group_component_dict.

        Compression disabled.
        """
        sg_compression_disabled = 'OS-myhost-SRP_1-DiamodNONE-os-iscsi-pg-CD'
        component_dict = self.utils.get_storage_group_component_dict(
            sg_compression_disabled)
        self.assertEqual('OS', component_dict['prefix'])
        self.assertEqual('myhost', component_dict['host'])
        self.assertEqual('SRP_1', component_dict['srp'])
        self.assertEqual('os-iscsi-pg', component_dict['portgroup'])
        self.assertEqual('DiamodNONE', component_dict['sloworkload'])
        self.assertEqual('-CD', component_dict['after_pg'])
        self.assertIsNone(component_dict['no_slo'])

    def test_get_storage_group_component_dict_replication_enabled(self):
        """Test for get_storage_group_component_dict.

        Replication enabled.
        """
        sg_slo_workload_rep = 'OS-myhost-SRP_1-DiamodOLTP-os-iscsi-pg-RE'
        component_dict = self.utils.get_storage_group_component_dict(
            sg_slo_workload_rep)
        self.assertEqual('OS', component_dict['prefix'])
        self.assertEqual('myhost', component_dict['host'])
        self.assertEqual('SRP_1', component_dict['srp'])
        self.assertEqual('os-iscsi-pg', component_dict['portgroup'])
        self.assertEqual('DiamodOLTP', component_dict['sloworkload'])
        self.assertEqual('-RE', component_dict['after_pg'])
        self.assertIsNone(component_dict['no_slo'])

    def test_get_storage_group_component_dict_slo_no_workload(self):
        """Test for get_storage_group_component_dict.

        SLO and no workload.
        """
        sg_slo_no_workload = 'OS-myhost-SRP_1-DiamodNONE-os-iscsi-pg'
        component_dict = self.utils.get_storage_group_component_dict(
            sg_slo_no_workload)
        self.assertEqual('OS', component_dict['prefix'])
        self.assertEqual('myhost', component_dict['host'])
        self.assertEqual('SRP_1', component_dict['srp'])
        self.assertEqual('os-iscsi-pg', component_dict['portgroup'])
        self.assertEqual('DiamodNONE', component_dict['sloworkload'])
        self.assertIsNone(component_dict['no_slo'])

    def test_get_storage_group_component_dict_dashes(self):
        """Test for get_storage_group_component_dict, dashes."""
        sg_host_with_dashes = (
            'OS-host-with-dashes-SRP_1-DiamodOLTP-myportgroup-RE')
        component_dict = self.utils.get_storage_group_component_dict(
            sg_host_with_dashes)
        self.assertEqual('host-with-dashes', component_dict['host'])
        self.assertEqual('OS', component_dict['prefix'])
        self.assertEqual('SRP_1', component_dict['srp'])
        self.assertEqual('DiamodOLTP', component_dict['sloworkload'])
        self.assertEqual('myportgroup', component_dict['portgroup'])
        self.assertEqual('-RE', component_dict['after_pg'])

    def test_delete_values_from_dict(self):
        """Test delete_values_from_dict"""
        delete_list = ['rdf_group_no', 'rep_mode', 'target_array_model',
                       'service_level', 'remote_array', 'target_device_id',
                       'replication_status', 'rdf_group_label']
        data_dict = self.utils.delete_values_from_dict(
            self.data.retype_metadata_dict, delete_list)
        self.assertEqual({'device_id': self.data.device_id}, data_dict)

    def test_update_values_in_dict(self):
        """Test delete_values_from_dict"""
        update_list = [('default_sg_name', 'source_sg_name'),
                       ('service_level', 'source_service_level')]

        update_dict = {'default_sg_name': 'default-sg',
                       'service_level': 'Diamond'}
        ret_dict = {'source_sg_name': 'default-sg',
                    'source_service_level': 'Diamond'}
        data_dict = self.utils.update_values_in_dict(
            update_dict, update_list)
        self.assertEqual(ret_dict, data_dict)

    def test_get_unique_device_ids_from_lists(self):
        list_a = ['00001', '00002', '00003']
        list_b = ['00002', '00003', '00004']
        unique_ids = self.utils.get_unique_device_ids_from_lists(
            list_a, list_b)
        self.assertEqual(['00004'], unique_ids)

    def test_update_payload_for_rdf_vol_create(self):
        payload = {
            'array': self.data.array,
            'editStorageGroupActionParam': {
                'expandStorageGroupParam': {
                    'addVolumeParam': {'create_new_volumes': 'False'}}}}

        updated_payload = self.utils.update_payload_for_rdf_vol_create(
            payload, self.data.remote_array, self.data.storagegroup_name_f)
        expected_payload = {
            'array': self.data.array,
            'editStorageGroupActionParam': {
                'expandStorageGroupParam': {
                    'addVolumeParam': {
                        'create_new_volumes': 'True',
                        'remoteSymmSGInfoParam': {
                            'force': 'true',
                            'remote_symmetrix_1_id': self.data.remote_array,
                            'remote_symmetrix_1_sgs': [
                                self.data.storagegroup_name_f]}}}}}
        self.assertEqual(expected_payload, updated_payload)

    def test_is_retype_supported(self):
        # Volume source type not replicated, target type Metro replicated,
        # volume is detached, host-assisted retype supported
        volume = self.data.test_volume
        volume.attach_status = 'detached'

        src_extra_specs = deepcopy(self.data.extra_specs)
        src_extra_specs['rep_mode'] = None

        tgt_extra_specs = deepcopy(self.data.rep_extra_specs)
        tgt_extra_specs['rep_mode'] = utils.REP_METRO

        rep_configs = self.data.multi_rep_config_list
        src_extra_specs[utils.REPLICATION_DEVICE_BACKEND_ID] = (
            self.data.rep_backend_id_sync)
        tgt_extra_specs[utils.REPLICATION_DEVICE_BACKEND_ID] = (
            self.data.rep_backend_id_metro)

        self.assertTrue(self.utils.is_retype_supported(
            volume, src_extra_specs, tgt_extra_specs, rep_configs))

        # Volume source type not replicated, target type Metro replicated,
        # volume is attached, host-assisted retype not supported
        volume.attach_status = 'attached'
        self.assertFalse(self.utils.is_retype_supported(
            volume, src_extra_specs, tgt_extra_specs, rep_configs))

        # Volume source type Async replicated, target type Metro replicated,
        # volume is attached, host-assisted retype not supported
        src_extra_specs['rep_mode'] = utils.REP_ASYNC
        self.assertFalse(self.utils.is_retype_supported(
            volume, src_extra_specs, tgt_extra_specs, rep_configs))

        # Volume source type Metro replicated, target type Metro replicated,
        # volume is attached, host-assisted retype supported
        src_extra_specs['rep_mode'] = utils.REP_METRO
        self.assertTrue(self.utils.is_retype_supported(
            volume, src_extra_specs, tgt_extra_specs, rep_configs))

    def test_validate_multiple_rep_device(self):
        self.utils.validate_multiple_rep_device(self.data.multi_rep_device)

    def test_validate_multiple_rep_device_non_unique_backend_id(self):
        rep_devices = deepcopy(self.data.multi_rep_device)
        rep_devices[0][utils.BACKEND_ID] = rep_devices[1][utils.BACKEND_ID]
        self.assertRaises(
            exception.InvalidConfigurationValue,
            self.utils.validate_multiple_rep_device,
            rep_devices)

    def test_validate_multiple_rep_device_promotion_start_backend_id(self):
        backend_id = utils.PMAX_FAILOVER_START_ARRAY_PROMOTION
        rep_devices = deepcopy(self.data.multi_rep_device)
        rep_devices[0][utils.BACKEND_ID] = backend_id
        self.assertRaises(
            exception.InvalidConfigurationValue,
            self.utils.validate_multiple_rep_device,
            rep_devices)

    def test_validate_multiple_rep_device_missing_backend_id(self):
        rep_devices = deepcopy(self.data.multi_rep_device)
        rep_devices[0].pop(utils.BACKEND_ID)
        self.assertRaises(
            exception.InvalidConfigurationValue,
            self.utils.validate_multiple_rep_device,
            rep_devices)

    def test_validate_multiple_rep_device_non_unique_rdf_label(self):
        rep_devices = deepcopy(self.data.multi_rep_device)
        rep_devices[0]['rdf_group_label'] = rep_devices[1]['rdf_group_label']
        self.assertRaises(
            exception.InvalidConfigurationValue,
            self.utils.validate_multiple_rep_device,
            rep_devices)

    def test_validate_multiple_rep_device_non_unique_rdf_modes(self):
        rep_devices = [self.data.rep_dev_1, deepcopy(self.data.rep_dev_2)]
        rep_devices[1]['mode'] = rep_devices[0]['mode']
        self.assertRaises(
            exception.InvalidConfigurationValue,
            self.utils.validate_multiple_rep_device,
            rep_devices)

    def test_validate_multiple_rep_device_defaulting_rdf_modes(self):
        rep_devices = [
            deepcopy(self.data.rep_dev_1), deepcopy(self.data.rep_dev_2)]
        rep_devices[0]['mode'] = ''
        rep_devices[1]['mode'] = 'testing'
        self.assertRaises(
            exception.InvalidConfigurationValue,
            self.utils.validate_multiple_rep_device,
            rep_devices)

    def test_validate_multiple_rep_device_multiple_targets(self):
        rep_devices = [self.data.rep_dev_1, deepcopy(self.data.rep_dev_2)]
        rep_devices[1]['target_device_id'] = 1234
        self.assertRaises(
            exception.InvalidConfigurationValue,
            self.utils.validate_multiple_rep_device,
            rep_devices)

    def test_validate_multiple_rep_device_length(self):
        rep_devices = [1, 2, 3, 4]
        self.assertRaises(
            exception.InvalidConfigurationValue,
            self.utils.validate_multiple_rep_device,
            rep_devices)

    def test_get_rep_config_single_rep(self):
        rep_configs = self.data.sync_rep_config_list
        rep_config = self.utils.get_rep_config('test', rep_configs)
        self.assertEqual(rep_config, rep_configs[0])

    def test_get_rep_config_multi_rep(self):
        rep_configs = self.data.multi_rep_config_list
        backend_id = rep_configs[0][utils.BACKEND_ID]
        rep_device = self.utils.get_rep_config(backend_id, rep_configs)
        self.assertEqual(rep_configs[0], rep_device)

    def test_get_rep_config_fail_non_legacy_backend_id_message(self):
        rep_configs = self.data.multi_rep_config_list
        backend_id = 'invalid_backend_id'
        try:
            self.utils.get_rep_config(backend_id, rep_configs)
        except exception.InvalidInput as e:
            expected_str = 'Could not find replication_device. Legacy'
            excep_msg = str(e)
            self.assertNotIn(expected_str, excep_msg)

    def test_get_rep_config_fail_legacy_backend_id_message(self):
        rep_configs = self.data.multi_rep_config_list
        backend_id = utils.BACKEND_ID_LEGACY_REP
        try:
            self.utils.get_rep_config(backend_id, rep_configs)
        except exception.InvalidInput as e:
            expected_str = 'Could not find replication_device. Legacy'
            excep_msg = str(e)
            self.assertIn(expected_str, excep_msg)

    def test_get_rep_config_promotion_stats(self):
        rep_configs = self.data.multi_rep_config_list
        backend_id = 'testing'
        rep_device = self.utils.get_rep_config(backend_id, rep_configs, True)
        self.assertEqual(rep_configs[0], rep_device)

    def test_get_replication_targets(self):
        rep_targets_expected = [self.data.remote_array]
        rep_configs = self.data.multi_rep_config_list
        rep_targets_actual = self.utils.get_replication_targets(rep_configs)
        self.assertEqual(rep_targets_expected, rep_targets_actual)

    def test_validate_failover_request_success(self):
        is_failed_over = False
        is_promoted = False
        failover_backend_id = self.data.rep_backend_id_sync
        rep_configs = self.data.multi_rep_config_list
        primary_array = self.data.array
        array_list = [self.data.array]
        is_valid, msg = self.utils.validate_failover_request(
            is_failed_over, failover_backend_id, rep_configs,
            primary_array, array_list, is_promoted)
        self.assertTrue(is_valid)
        self.assertEqual("", msg)

    def test_validate_failover_request_already_failed_over(self):
        is_failed_over = True
        is_promoted = False
        failover_backend_id = self.data.rep_backend_id_sync
        rep_configs = self.data.multi_rep_config_list
        primary_array = self.data.array
        array_list = [self.data.array]
        is_valid, msg = self.utils.validate_failover_request(
            is_failed_over, failover_backend_id, rep_configs,
            primary_array, array_list, is_promoted)
        self.assertFalse(is_valid)
        expected_msg = ('Cannot failover, the backend is already in a failed '
                        'over state, if you meant to failback, please add '
                        '--backend_id default to the command.')
        self.assertEqual(expected_msg, msg)

    def test_validate_failover_request_failback_missing_array(self):
        is_failed_over = True
        is_promoted = False
        failover_backend_id = 'default'
        rep_configs = self.data.multi_rep_config_list
        primary_array = self.data.array
        array_list = [self.data.remote_array]
        is_valid, msg = self.utils.validate_failover_request(
            is_failed_over, failover_backend_id, rep_configs,
            primary_array, array_list, is_promoted)
        self.assertFalse(is_valid)
        expected_msg = ('Cannot failback, the configured primary array is '
                        'not currently available to perform failback to. '
                        'Please ensure array %s is visible in '
                        'Unisphere.') % primary_array
        self.assertEqual(expected_msg, msg)

    def test_validate_failover_request_promotion_finalize(self):
        is_failed_over = True
        is_promoted = True
        failover_backend_id = utils.PMAX_FAILOVER_START_ARRAY_PROMOTION
        rep_configs = self.data.multi_rep_config_list
        primary_array = self.data.array
        array_list = [self.data.array]
        is_valid, msg = self.utils.validate_failover_request(
            is_failed_over, failover_backend_id, rep_configs,
            primary_array, array_list, is_promoted)
        self.assertFalse(is_valid)
        expected_msg = ('Failover promotion currently in progress, please '
                        'finish the promotion process and issue a failover '
                        'using the "default" backend_id to complete this '
                        'process.')
        self.assertEqual(expected_msg, msg)

    def test_validate_failover_request_invalid_failback(self):
        is_failed_over = False
        is_promoted = False
        failover_backend_id = 'default'
        rep_configs = self.data.multi_rep_config_list
        primary_array = self.data.array
        array_list = [self.data.array]
        is_valid, msg = self.utils.validate_failover_request(
            is_failed_over, failover_backend_id, rep_configs,
            primary_array, array_list, is_promoted)
        self.assertFalse(is_valid)
        expected_msg = ('Cannot failback, backend is not in a failed over '
                        'state. If you meant to failover, please either omit '
                        'the --backend_id parameter or use the --backend_id '
                        'parameter with a valid backend id.')
        self.assertEqual(expected_msg, msg)

    def test_validate_replication_group_config_success(self):
        rep_configs = deepcopy(self.data.multi_rep_config_list)
        extra_specs = deepcopy(
            self.data.vol_type_extra_specs_rep_enabled_backend_id_sync)
        extra_specs[utils.REPLICATION_DEVICE_BACKEND_ID] = (
            self.data.rep_backend_id_sync)
        self.utils.validate_replication_group_config(
            rep_configs, [extra_specs])

    def test_validate_replication_group_config_no_rep_configured(self):
        rep_configs = None
        extra_specs_list = [
            self.data.vol_type_extra_specs_rep_enabled_backend_id_sync]
        self.assertRaises(exception.InvalidInput,
                          self.utils.validate_replication_group_config,
                          rep_configs, extra_specs_list)
        try:
            self.utils.validate_replication_group_config(
                rep_configs, extra_specs_list)
        except exception.InvalidInput as e:
            expected_msg = (
                'Invalid input received: No replication devices are defined '
                'in cinder.conf, can not enable volume group replication.')
            self.assertEqual(expected_msg, e.msg)

    def test_validate_replication_group_config_vol_type_not_rep_enabled(self):
        rep_configs = self.data.multi_rep_config_list
        extra_specs_list = [self.data.vol_type_extra_specs]
        self.assertRaises(exception.InvalidInput,
                          self.utils.validate_replication_group_config,
                          rep_configs, extra_specs_list)
        try:
            self.utils.validate_replication_group_config(
                rep_configs, extra_specs_list)
        except exception.InvalidInput as e:
            expected_msg = (
                'Invalid input received: Replication is not enabled for a '
                'Volume Type, all Volume Types in a replication enabled '
                'Volume Group must have replication enabled.')
            self.assertEqual(expected_msg, e.msg)

    def test_validate_replication_group_config_cant_get_rep_config(self):
        rep_configs = self.data.multi_rep_config_list
        vt_extra_specs = (
            self.data.vol_type_extra_specs_rep_enabled_backend_id_sync)
        vt_extra_specs[utils.REPLICATION_DEVICE_BACKEND_ID] = 'invalid'
        extra_specs_list = [vt_extra_specs]
        self.assertRaises(exception.InvalidInput,
                          self.utils.validate_replication_group_config,
                          rep_configs, extra_specs_list)
        try:
            self.utils.validate_replication_group_config(
                rep_configs, extra_specs_list)
        except exception.InvalidInput as e:
            expected_msg = (
                'Invalid input received: Unable to determine which '
                'rep_device to use from cinder.conf. Could not validate '
                'volume types being added to group.')
            self.assertEqual(expected_msg, e.msg)

    def test_validate_replication_group_config_non_sync_mode(self):
        rep_configs = self.data.multi_rep_config_list
        extra_specs_list = [
            self.data.vol_type_extra_specs_rep_enabled_backend_id_async]
        self.assertRaises(exception.InvalidInput,
                          self.utils.validate_replication_group_config,
                          rep_configs, extra_specs_list)
        try:
            self.utils.validate_replication_group_config(
                rep_configs, extra_specs_list)
        except exception.InvalidInput as e:
            expected_msg = (
                'Invalid input received: Replication for Volume Type is not '
                'set to Synchronous. Only Synchronous can be used with '
                'replication groups')
            self.assertEqual(expected_msg, e.msg)

    @mock.patch.object(utils.PowerMaxUtils, 'get_rep_config')
    def test_validate_replication_group_config_multiple_rep_backend_ids(
            self, mck_get):
        side_effect_list = [
            self.data.rep_config_sync, self.data.rep_config_sync_2]
        mck_get.side_effect = side_effect_list
        rep_configs = self.data.multi_rep_config_list
        ex_specs_1 = deepcopy(
            self.data.vol_type_extra_specs_rep_enabled_backend_id_sync)
        ex_specs_2 = deepcopy(
            self.data.vol_type_extra_specs_rep_enabled_backend_id_sync_2)
        extra_specs_list = [ex_specs_1, ex_specs_2]
        self.assertRaises(exception.InvalidInput,
                          self.utils.validate_replication_group_config,
                          rep_configs, extra_specs_list)
        mck_get.side_effect = side_effect_list
        try:
            self.utils.validate_replication_group_config(
                rep_configs, extra_specs_list)
        except exception.InvalidInput as e:
            expected_msg = (
                'Invalid input received: Multiple replication backend ids '
                'detected please ensure only a single replication device '
                '(backend_id) is used for all Volume Types in a Volume '
                'Group.')
            self.assertEqual(expected_msg, e.msg)

    def test_validate_non_replication_group_config_success(self):
        extra_specs_list = [
            self.data.vol_type_extra_specs]
        self.utils.validate_non_replication_group_config(extra_specs_list)

    def test_validate_non_replication_group_config_failure(self):
        extra_specs = {'pool_name': u'Diamond+DSS+SRP_1+000197800123',
                       utils.IS_RE: '<is> True'}
        self.assertRaises(exception.InvalidInput,
                          self.utils.validate_non_replication_group_config,
                          [extra_specs])
        try:
            self.utils.validate_non_replication_group_config([extra_specs])
        except exception.InvalidInput as e:
            expected_msg = (
                'Invalid input received: Replication is enabled in one or '
                'more of the Volume Types being added to new Volume Group '
                'but the Volume Group is not replication enabled. Please '
                'enable replication in the Volume Group or select only '
                'non-replicated Volume Types.')
            self.assertEqual(expected_msg, e.msg)

    def test_get_migration_delete_extra_specs_replicated(self):
        volume = self.data.test_volume
        metadata = deepcopy(self.data.volume_metadata)
        metadata[utils.IS_RE_CAMEL] = 'True'
        metadata['ReplicationMode'] = utils.REP_SYNC
        metadata['RDFG-Label'] = self.data.rdf_group_name_1
        volume.metadata = metadata
        extra_specs = deepcopy(self.data.extra_specs)
        rep_configs = self.data.multi_rep_config_list
        updated_extra_specs = self.utils.get_migration_delete_extra_specs(
            volume, extra_specs, rep_configs)
        ref_extra_specs = deepcopy(self.data.extra_specs)
        ref_extra_specs[utils.IS_RE] = True
        ref_extra_specs[utils.REP_MODE] = utils.REP_SYNC
        ref_extra_specs[utils.REP_CONFIG] = self.data.rep_config_sync
        ref_extra_specs[utils.REPLICATION_DEVICE_BACKEND_ID] = (
            self.data.rep_backend_id_sync)
        self.assertEqual(ref_extra_specs, updated_extra_specs)

    def test_get_migration_delete_extra_specs_non_replicated(self):
        volume = self.data.test_volume
        volume.metadata = self.data.volume_metadata
        extra_specs = deepcopy(self.data.extra_specs)
        extra_specs[utils.IS_RE] = True
        updated_extra_specs = self.utils.get_migration_delete_extra_specs(
            volume, extra_specs, None)
        self.assertEqual(self.data.extra_specs, updated_extra_specs)

    def test_version_meet_req_true(self):
        version = '9.1.0.14'
        minimum_version = '9.1.0.5'
        self.assertTrue(
            self.utils.version_meet_req(version, minimum_version))

    def test_version_meet_req_false(self):
        version = '9.1.0.3'
        minimum_version = '9.1.0.5'
        self.assertFalse(
            self.utils.version_meet_req(version, minimum_version))

    def test_version_meet_req_major_true(self):
        version = '9.2.0.1'
        minimum_version = '9.1.0.5'
        self.assertTrue(
            self.utils.version_meet_req(version, minimum_version))

    def test_parse_specs_from_pool_name_workload_included(self):
        pool_name = self.data.vol_type_extra_specs.get('pool_name')
        array_id, srp, service_level, workload = (
            self.utils.parse_specs_from_pool_name(pool_name))
        pool_details = pool_name.split('+')
        self.assertEqual(array_id, pool_details[3])
        self.assertEqual(srp, pool_details[2])
        self.assertEqual(workload, pool_details[1])
        self.assertEqual(service_level, pool_details[0])

    def test_parse_specs_from_pool_name_workload_not_included(self):
        pool_name = (
            self.data.vol_type_extra_specs_next_gen_pool.get('pool_name'))
        array_id, srp, service_level, workload = (
            self.utils.parse_specs_from_pool_name(pool_name))

        pool_details = pool_name.split('+')
        self.assertEqual(array_id, pool_details[2])
        self.assertEqual(srp, pool_details[1])
        self.assertEqual(service_level, pool_details[0])
        self.assertEqual(workload, str())

    def test_parse_specs_from_pool_name_invalid_pool(self):
        pool_name = 'This+Is+An+Invalid+Pool'
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.utils.parse_specs_from_pool_name, pool_name)

    def test_parse_specs_from_pool_name_no_pool(self):
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.utils.parse_specs_from_pool_name, '')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class FakeVolumeAPI(object):
    def __init__(self, expected_spec, test_inst):
        self.expected_spec = expected_spec.copy()
        self.test_inst = test_inst

    def create_volume(self, ctxt, volume, host,
                      request_spec, filter_properties,
                      allow_reschedule=True,
                      snapshot_id=None, image_id=None,
                      source_volid=None,
                      source_replicaid=None):

        self.test_inst.assertEqual(self.expected_spec, request_spec)
        self.test_inst.assertEqual(request_spec['source_volid'], source_volid)
        self.test_inst.assertEqual(request_spec['snapshot_id'], snapshot_id)
        self.test_inst.assertEqual(request_spec['image_id'], image_id)


class FakeSchedulerRpcAPI(object):
    def __init__(self, expected_spec, test_inst):
        self.expected_spec = expected_spec.copy()
        self.test_inst = test_inst

    def create_volume(self, ctxt, volume, snapshot_id=None, image_id=None,
                      request_spec=None, filter_properties=None,
                      backup_id=None):

        self.test_inst.assertEqual(self.expected_spec, request_spec)

    def manage_existing(self, context, volume, request_spec=None):
        self.test_inst.assertEqual(self.expected_spec, request_spec)


class FakeDb(object):

    def volume_get(self, *args, **kwargs):
        return {'host': 'barf'}

    def volume_update(self, *args, **kwargs):
        return {'host': 'farb'}

    def snapshot_get(self, *args, **kwargs):
        return {'volume_id': 1}

    def consistencygroup_get(self, *args, **kwargs):
        return {'consistencygroup_id': 1}



# Copyright (c) 2016 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from unittest import mock

import ddt
import iso8601
from oslo_utils import versionutils

from cinder.api import extensions
from cinder.api import microversions as mv
from cinder.api.openstack import api_version_request as api_version
from cinder.api.v3 import clusters
from cinder import context
from cinder import exception
from cinder.tests.unit import fake_cluster
from cinder.tests.unit import test


CLUSTERS = [
    fake_cluster.fake_db_cluster(
        id=1,
        replication_status='error',
        frozen=False,
        active_backend_id='replication1',
        last_heartbeat=datetime.datetime(2016, 6, 1, 2, 46, 28),
        updated_at=datetime.datetime(2016, 6, 1, 2, 46, 28),
        created_at=datetime.datetime(2016, 6, 1, 2, 46, 28)),
    fake_cluster.fake_db_cluster(
        id=2, name='cluster2', num_hosts=2, num_down_hosts=1, disabled=True,
        replication_status='error',
        frozen=True,
        active_backend_id='replication2',
        updated_at=datetime.datetime(2016, 6, 1, 1, 46, 28),
        created_at=datetime.datetime(2016, 6, 1, 1, 46, 28))
]

CLUSTERS_ORM = [fake_cluster.fake_cluster_orm(**kwargs) for kwargs in CLUSTERS]

EXPECTED = [{'created_at': datetime.datetime(2016, 6, 1, 2, 46, 28),
             'disabled_reason': None,
             'last_heartbeat': datetime.datetime(2016, 6, 1, 2, 46, 28),
             'name': 'cluster_name',
             'binary': 'cinder-volume',
             'num_down_hosts': 0,
             'num_hosts': 0,
             'state': 'up',
             'status': 'enabled',
             'replication_status': 'error',
             'frozen': False,
             'active_backend_id': 'replication1',
             'updated_at': datetime.datetime(2016, 6, 1, 2, 46, 28)},
            {'created_at': datetime.datetime(2016, 6, 1, 1, 46, 28),
             'disabled_reason': None,
             'last_heartbeat': '',
             'name': 'cluster2',
             'binary': 'cinder-volume',
             'num_down_hosts': 1,
             'num_hosts': 2,
             'state': 'down',
             'status': 'disabled',
             'replication_status': 'error',
             'frozen': True,
             'active_backend_id': 'replication2',
             'updated_at': datetime.datetime(2016, 6, 1, 1, 46, 28)}]


class FakeRequest(object):
    def __init__(self, is_admin=True, version=mv.CLUSTER_SUPPORT, **kwargs):
        self.GET = kwargs
        self.headers = {'OpenStack-API-Version': 'volume ' + version}
        self.api_version_request = api_version.APIVersionRequest(version)
        self.environ = {
            'cinder.context': context.RequestContext(user_id=None,
                                                     project_id=None,
                                                     is_admin=is_admin,
                                                     read_deleted='no',
                                                     overwrite=False)
        }


def fake_utcnow(with_timezone=False):
    tzinfo = iso8601.UTC if with_timezone else None
    return datetime.datetime(2016, 6, 1, 2, 46, 30, tzinfo=tzinfo)


@ddt.ddt
@mock.patch('oslo_utils.timeutils.utcnow', fake_utcnow)
class ClustersTestCase(test.TestCase):
    """Test Case for Clusters."""
    LIST_FILTERS = ({}, {'is_up': True}, {'disabled': False}, {'num_hosts': 2},
                    {'num_down_hosts': 1}, {'binary': 'cinder-volume'},
                    {'is_up': True, 'disabled': False, 'num_hosts': 2,
                     'num_down_hosts': 1, 'binary': 'cinder-volume'})

    REPLICATION_FILTERS = ({'replication_status': 'error'}, {'frozen': True},
                           {'active_backend_id': 'replication'})

    def _get_expected(self,
                      version=mv.get_prior_version(mv.REPLICATION_CLUSTER)):
        if (versionutils.convert_version_to_tuple(version) >=
                versionutils.convert_version_to_tuple(mv.REPLICATION_CLUSTER)):
            return EXPECTED

        expect = []
        for cluster in EXPECTED:
            cluster = cluster.copy()
            for key in ('replication_status', 'frozen', 'active_backend_id'):
                cluster.pop(key)
            expect.append(cluster)
        return expect

    def setUp(self):
        super(ClustersTestCase, self).setUp()

        self.context = context.get_admin_context()
        self.ext_mgr = extensions.ExtensionManager()
        self.ext_mgr.extensions = {}
        self.controller = clusters.ClusterController(self.ext_mgr)

    @mock.patch('cinder.db.cluster_get_all', return_value=CLUSTERS_ORM)
    def _test_list(self, get_all_mock, detailed, filters=None, expected=None,
                   version=mv.get_prior_version(mv.REPLICATION_CLUSTER)):
        filters = filters or {}
        req = FakeRequest(version=version, **filters)
        method = getattr(self.controller, 'detail' if detailed else 'index')
        clusters = method(req)

        filters = filters.copy()
        filters.setdefault('is_up', None)
        filters.setdefault('read_deleted', 'no')
        self.assertEqual(expected, clusters)
        get_all_mock.assert_called_once_with(
            req.environ['cinder.context'],
            get_services=False,
            services_summary=detailed,
            **filters)

    @ddt.data(*LIST_FILTERS)
    def test_index_detail(self, filters):
        """Verify that we get all clusters with detailed data."""
        expected = {'clusters': self._get_expected()}
        self._test_list(detailed=True, filters=filters, expected=expected)

    @ddt.data(*LIST_FILTERS)
    def test_index_summary(self, filters):
        """Verify that we get all clusters with summary data."""
        expected = {'clusters': [{'name': 'cluster_name',
                                  'binary': 'cinder-volume',
                                  'state': 'up',
                                  'status': 'enabled'},
                                 {'name': 'cluster2',
                                  'binary': 'cinder-volume',
                                  'state': 'down',
                                  'status': 'disabled'}]}
        self._test_list(detailed=False, filters=filters, expected=expected)

    @ddt.data(*REPLICATION_FILTERS)
    def test_index_detail_fail_old(self, filters):
        self.assertRaises(exception.InvalidInput, self._test_list,
                          detailed=True, filters=filters)

    @ddt.data(*REPLICATION_FILTERS)
    def test_index_summary_fail_old(self, filters):
        self.assertRaises(exception.InvalidInput, self._test_list,
                          detailed=False, filters=filters)

    @ddt.data(True, False)
    def test_index_unauthorized(self, detailed):
        """Verify that unauthorized user can't list clusters."""
        self.assertRaises(exception.PolicyNotAuthorized,
                          self._test_list, detailed=detailed,
                          filters={'is_admin': False})

    @ddt.data(True, False)
    def test_index_wrong_version(self, detailed):
        """Verify the wrong version so that user can't list clusters."""
        self.assertRaises(exception.VersionNotFoundForAPIMethod,
                          self._test_list, detailed=detailed,
                          version=mv.get_prior_version(mv.CLUSTER_SUPPORT))

    @ddt.data(*REPLICATION_FILTERS)
    def test_index_detail_replication_new_fields(self, filters):
        expected = {'clusters': self._get_expected(mv.REPLICATION_CLUSTER)}
        self._test_list(detailed=True, filters=filters, expected=expected,
                        version=mv.REPLICATION_CLUSTER)

    @ddt.data(*REPLICATION_FILTERS)
    def test_index_summary_replication_new_fields(self, filters):
        expected = {'clusters': [{'name': 'cluster_name',
                                  'binary': 'cinder-volume',
                                  'state': 'up',
                                  'replication_status': 'error',
                                  'status': 'enabled'},
                                 {'name': 'cluster2',
                                  'binary': 'cinder-volume',
                                  'state': 'down',
                                  'replication_status': 'error',
                                  'status': 'disabled'}]}
        self._test_list(detailed=False, filters=filters, expected=expected,
                        version=mv.REPLICATION_CLUSTER)

    @mock.patch('cinder.db.sqlalchemy.api.cluster_get',
                return_value=CLUSTERS_ORM[0])
    def test_show(self, get_mock):
        req = FakeRequest()
        expected = {'cluster': self._get_expected()[0]}
        cluster = self.controller.show(req, 'cluster_name',
                                       'cinder-volume')
        self.assertEqual(expected, cluster)
        get_mock.assert_called_once_with(
            req.environ['cinder.context'],
            None,
            services_summary=True,
            name='cluster_name',
            binary='cinder-volume')

    def test_show_unauthorized(self):
        req = FakeRequest(is_admin=False)
        self.assertRaises(exception.PolicyNotAuthorized,
                          self.controller.show, req, 'name')

    def test_show_wrong_version(self):
        req = FakeRequest(version=mv.get_prior_version(mv.CLUSTER_SUPPORT))
        self.assertRaises(exception.VersionNotFoundForAPIMethod,
                          self.controller.show, req, 'name')

    @mock.patch('cinder.db.sqlalchemy.api.cluster_update')
    @mock.patch('cinder.db.sqlalchemy.api.cluster_get',
                return_value=CLUSTERS_ORM[1])
    def test_update_enable(self, get_mock, update_mock):
        req = FakeRequest()
        expected = {'cluster': {'name': 'cluster2',
                                'binary': 'cinder-volume',
                                'state': 'down',
                                'status': 'enabled',
                                'disabled_reason': None}}
        res = self.controller.update(req, 'enable',
                                     body={'name': 'cluster_name',
                                           'binary': 'cinder-volume'})
        self.assertEqual(expected, res)
        ctxt = req.environ['cinder.context']
        get_mock.assert_called_once_with(ctxt,
                                         None, binary='cinder-volume',
                                         name='cluster_name')
        update_mock.assert_called_once_with(ctxt, get_mock.return_value.id,
                                            {'disabled': False,
                                             'disabled_reason': None})

    @mock.patch('cinder.db.sqlalchemy.api.cluster_update')
    @mock.patch('cinder.db.sqlalchemy.api.cluster_get',
                return_value=CLUSTERS_ORM[0])
    def test_update_disable(self, get_mock, update_mock):
        req = FakeRequest()
        disabled_reason = 'For testing'
        expected = {'cluster': {'name': 'cluster_name',
                                'state': 'up',
                                'binary': 'cinder-volume',
                                'status': 'disabled',
                                'disabled_reason': disabled_reason}}
        res = self.controller.update(req, 'disable',
                                     body={'name': 'cluster_name',
                                           'binary': 'cinder-volume',
                                           'disabled_reason': disabled_reason})
        self.assertEqual(expected, res)
        ctxt = req.environ['cinder.context']
        get_mock.assert_called_once_with(ctxt,
                                         None, binary='cinder-volume',
                                         name='cluster_name')
        update_mock.assert_called_once_with(
            ctxt, get_mock.return_value.id,
            {'disabled': True, 'disabled_reason': disabled_reason})

    def test_update_wrong_action(self):
        req = FakeRequest()
        self.assertRaises(exception.NotFound, self.controller.update,
                          req, 'action', body={'name': 'cluster_name'})

    @ddt.data('enable', 'disable')
    def test_update_missing_name(self, action):
        req = FakeRequest()
        self.assertRaises(exception.ValidationError, self.controller.update,
                          req, action, body={'binary': 'cinder-volume'})

    def test_update_with_binary_more_than_255_characters(self):
        req = FakeRequest()
        self.assertRaises(exception.ValidationError, self.controller.update,
                          req, 'enable', body={'name': 'cluster_name',
                                               'binary': 'a' * 256})

    def test_update_with_name_more_than_255_characters(self):
        req = FakeRequest()
        self.assertRaises(exception.ValidationError, self.controller.update,
                          req, 'enable', body={'name': 'a' * 256,
                                               'binary': 'cinder-volume'})

    @ddt.data('a' * 256, '   ')
    def test_update_wrong_disabled_reason(self, disabled_reason):
        req = FakeRequest()
        self.assertRaises(exception.ValidationError, self.controller.update,
                          req, 'disable',
                          body={'name': 'cluster_name',
                                'disabled_reason': disabled_reason})

    @ddt.data('enable', 'disable')
    def test_update_unauthorized(self, action):
        req = FakeRequest(is_admin=False)
        self.assertRaises(exception.PolicyNotAuthorized,
                          self.controller.update, req, action,
                          body={'name': 'fake_name'})

    @ddt.data('enable', 'disable')
    def test_update_wrong_version(self, action):
        req = FakeRequest(version=mv.get_prior_version(mv.CLUSTER_SUPPORT))
        self.assertRaises(exception.VersionNotFoundForAPIMethod,
                          self.controller.update, req, action, {})



#    Copyright (c) 2020 Open-E, Inc.
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from oslo_utils import units as o_units

from cinder import context
from cinder import exception
from cinder.tests.unit import test
from cinder.volume.drivers.open_e.jovian_common import exception as jexc
from cinder.volume.drivers.open_e.jovian_common import jdss_common as jcom
from cinder.volume.drivers.open_e.jovian_common import rest

UUID_1 = '12345678-1234-1234-1234-000000000001'
UUID_2 = '12345678-1234-1234-1234-000000000002'
UUID_3 = '12345678-1234-1234-1234-000000000003'

CONFIG_OK = {
    'san_hosts': ['192.168.0.2'],
    'san_api_port': 82,
    'driver_use_ssl': 'true',
    'jovian_rest_send_repeats': 3,
    'jovian_recovery_delay': 60,
    'san_login': 'admin',
    'san_password': 'password',
    'jovian_ignore_tpath': [],
    'target_port': 3260,
    'jovian_pool': 'Pool-0',
    'iscsi_target_prefix': 'iqn.2020-04.com.open-e.cinder:',
    'chap_password_len': 12,
    'san_thin_provision': False,
    'jovian_block_size': '128K'
}


class TestOpenEJovianRESTAPI(test.TestCase):

    def get_rest(self, config):
        ctx = context.get_admin_context()

        cfg = mock.Mock()
        cfg.append_config_values.return_value = None
        cfg.safe_get = lambda val: config[val]
        cfg.get = lambda val, default: config.get(val, default)
        jdssr = rest.JovianRESTAPI(config)
        jdssr.rproxy = mock.Mock()
        return jdssr, ctx

    def test_get_active_host(self):

        jrest, ctx = self.get_rest(CONFIG_OK)

        jrest.rproxy.get_active_host.return_value = "test_data"

        ret = jrest.get_active_host()

        self.assertEqual("test_data", ret)

    def test_is_pool_exists(self):
        jrest, ctx = self.get_rest(CONFIG_OK)
        resp = {'code': 200,
                'error': None}

        jrest.rproxy.pool_request.return_value = resp
        self.assertTrue(jrest.is_pool_exists())

        err = {'errorid': 12}
        resp = {'code': 404,
                'error': err}
        jrest.rproxy.pool_request.return_value = resp
        self.assertFalse(jrest.is_pool_exists())

        pool_request_expected = [
            mock.call('GET', ''),
            mock.call('GET', '')]

        jrest.rproxy.pool_request.assert_has_calls(pool_request_expected)

    def get_iface_info(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        resp = {
            'code': 200,
            'error': None}
        jrest.rproxy.pool_request.return_value = resp
        self.assertTrue(jrest.is_pool_exists())

    def test_get_luns(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        resp = {'data': [{
                'vscan': None,
                'full_name': 'Pool-0/' + UUID_1,
                'userrefs': None,
                'primarycache': 'all',
                'logbias': 'latency',
                'creation': '1591543140',
                'sync': 'always',
                'is_clone': False,
                'dedup': 'off',
                'sharenfs': None,
                'receive_resume_token': None,
                'volsize': '1073741824'}],
                'error': None,
                'code': 200}
        jrest.rproxy.pool_request.return_value = resp
        self.assertEqual(resp['data'], jrest.get_luns())

        err = {'errorid': 12, 'message': 'test failure'}
        resp = {'code': 404,
                'data': None,
                'error': err}
        jrest.rproxy.pool_request.return_value = resp
        self.assertRaises(jexc.JDSSRESTException, jrest.get_luns)

        get_luns_expected = [
            mock.call('GET', "/volumes"),
            mock.call('GET', "/volumes")]

        jrest.rproxy.pool_request.assert_has_calls(get_luns_expected)

    def test_create_lun(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        resp = {'data': {
                'vscan': None,
                'full_name': 'Pool-0/' + jcom.vname(UUID_1),
                'userrefs': None,
                'primarycache': 'all',
                'logbias': 'latency',
                'creation': '1591543140',
                'sync': 'always',
                'is_clone': False,
                'dedup': 'off',
                'sharenfs': None,
                'receive_resume_token': None,
                'volsize': '1073741824'},
                'error': None,
                'code': 200}

        jbody = {
            'name': jcom.vname(UUID_1),
            'size': "1073741824",
            'sparse': False
        }

        jbody_sparse = {
            'name': jcom.vname(UUID_1),
            'size': "1073741824",
            'sparse': True
        }

        jrest.rproxy.pool_request.return_value = resp
        self.assertIsNone(jrest.create_lun(jcom.vname(UUID_1), o_units.Gi))

        err = {'errno': '5', 'message': 'test failure'}
        resp = {'code': 404,
                'data': None,
                'error': err}
        jrest.rproxy.pool_request.return_value = resp
        self.assertRaises(jexc.JDSSRESTException,
                          jrest.create_lun,
                          jcom.vname(UUID_1),
                          o_units.Gi,
                          sparse=True)

        addr = "/volumes"
        create_lun_expected = [
            mock.call('POST', addr, json_data=jbody),
            mock.call('POST', addr, json_data=jbody_sparse)]

        jrest.rproxy.pool_request.assert_has_calls(create_lun_expected)

    def test_extend_lun(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        resp = {'data': None,
                'error': None,
                'code': 201}

        jbody = {
            'size': "2147483648",
        }

        jrest.rproxy.pool_request.return_value = resp
        self.assertIsNone(jrest.extend_lun(jcom.vname(UUID_1), 2 * o_units.Gi))

        err = {'message': 'test failure'}
        resp = {'code': 500,
                'data': None,
                'error': err}
        jrest.rproxy.pool_request.return_value = resp
        self.assertRaises(jexc.JDSSRESTException,
                          jrest.extend_lun,
                          jcom.vname(UUID_1),
                          2 * o_units.Gi)

        addr = "/volumes/" + jcom.vname(UUID_1)
        create_lun_expected = [
            mock.call('PUT', addr, json_data=jbody),
            mock.call('PUT', addr, json_data=jbody)]

        jrest.rproxy.pool_request.assert_has_calls(create_lun_expected)

    def test_is_lun(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        resp = {'data': {
                "vscan": None,
                "full_name": "Pool-0/" + jcom.vname(UUID_1),
                "userrefs": None,
                "primarycache": "all",
                "logbias": "latency",
                "creation": "1591543140",
                "sync": "always",
                "is_clone": False,
                "dedup": "off",
                "sharenfs": None,
                "receive_resume_token": None,
                "volsize": "1073741824"},
                'error': None,
                'code': 200}

        jrest.rproxy.pool_request.return_value = resp
        self.assertTrue(jrest.is_lun(jcom.vname(UUID_1)))

        err = {'errno': 1,
               'message': ('Zfs resource: Pool-0/' + jcom.vname(UUID_1) +
                           ' not found in this collection.')}
        resp = {'code': 500,
                'data': None,
                'error': err}

        jrest.rproxy.pool_request.return_value = resp
        self.assertEqual(False, jrest.is_lun(jcom.vname(UUID_1)))

        jrest.rproxy.pool_request.side_effect = (
            jexc.JDSSRESTProxyException(host='test_host', reason='test'))

        self.assertRaises(jexc.JDSSRESTProxyException,
                          jrest.is_lun,
                          'v_' + UUID_1)

    def test_get_lun(self):
        jrest, ctx = self.get_rest(CONFIG_OK)
        resp = {'data': {"vscan": None,
                         "full_name": "Pool-0/v_" + UUID_1,
                         "userrefs": None,
                         "primarycache": "all",
                         "logbias": "latency",
                         "creation": "1591543140",
                         "sync": "always",
                         "is_clone": False,
                         "dedup": "off",
                         "sharenfs": None,
                         "receive_resume_token": None,
                         "volsize": "1073741824"},
                'error': None,
                'code': 200}

        jrest.rproxy.pool_request.return_value = resp
        self.assertEqual(resp['data'], jrest.get_lun('v_' + UUID_1))

        err = {'errno': 1,
               'message': ('Zfs resource: Pool-0/v_' + UUID_1 +
                           ' not found in this collection.')}
        resp = {'code': 500,
                'data': None,
                'error': err}

        jrest.rproxy.pool_request.return_value = resp
        self.assertRaises(jexc.JDSSResourceNotFoundException,
                          jrest.get_lun,
                          'v_' + UUID_1)

        jrest.rproxy.pool_request.return_value = resp
        self.assertRaises(jexc.JDSSResourceNotFoundException,
                          jrest.get_lun,
                          'v_' + UUID_1)

        err = {'errno': 10,
               'message': ('Test error')}
        resp = {'code': 500,
                'data': None,
                'error': err}

        jrest.rproxy.pool_request.return_value = resp
        self.assertRaises(jexc.JDSSException, jrest.get_lun, 'v_' + UUID_1)

    def test_modify_lun(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        resp = {'data': None,
                'error': None,
                'code': 201}
        req = {'name': 'v_' + UUID_2}

        jrest.rproxy.pool_request.return_value = resp
        self.assertIsNone(jrest.modify_lun('v_' + UUID_1, prop=req))

        err = {'errno': 1,
               'message': ('Zfs resource: Pool-0/v_' + UUID_1 +
                           ' not found in this collection.')}
        resp = {'code': 500,
                'data': None,
                'error': err}

        jrest.rproxy.pool_request.return_value = resp
        self.assertRaises(jexc.JDSSResourceNotFoundException,
                          jrest.modify_lun,
                          'v_' + UUID_1,
                          prop=req)

        err = {'errno': 10,
               'message': ('Test error')}
        resp = {'code': 500,
                'data': None,
                'error': err}

        jrest.rproxy.pool_request.return_value = resp
        self.assertRaises(jexc.JDSSException,
                          jrest.modify_lun,
                          'v_' + UUID_1,
                          prop=req)

        addr = "/volumes/v_" + UUID_1
        modify_lun_expected = [
            mock.call('PUT', addr, json_data=req),
            mock.call('PUT', addr, json_data=req),
            mock.call('PUT', addr, json_data=req)]

        jrest.rproxy.pool_request.assert_has_calls(modify_lun_expected)

    def test_make_readonly_lun(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        resp = {'data': None,
                'error': None,
                'code': 201}
        req = {'property_name': 'readonly', 'property_value': 'on'}

        jrest.rproxy.pool_request.return_value = resp
        self.assertIsNone(jrest.modify_lun('v_' + UUID_1, prop=req))

        addr = "/volumes/v_" + UUID_1
        modify_lun_expected = [mock.call('PUT', addr, json_data=req)]

        jrest.rproxy.pool_request.assert_has_calls(modify_lun_expected)

    def test_delete_lun(self):

        jrest, ctx = self.get_rest(CONFIG_OK)

        # Delete OK
        resp = {'data': None,
                'error': None,
                'code': 204}

        jrest.rproxy.pool_request.return_value = resp
        self.assertIsNone(jrest.delete_lun('v_' + UUID_1))
        addr = "/volumes/v_" + UUID_1
        delete_lun_expected = [mock.call('DELETE', addr)]
        jrest.rproxy.pool_request.assert_has_calls(delete_lun_expected)
        # No volume to delete
        err = {'errno': 1,
               'message': ('Zfs resource: Pool-0/v_' + UUID_1 +
                           ' not found in this collection.')}
        resp = {'code': 500,
                'data': None,
                'error': err}

        jrest.rproxy.pool_request.return_value = resp
        self.assertIsNone(jrest.delete_lun('v_' + UUID_1))

        delete_lun_expected += [mock.call('DELETE', addr)]

        jrest.rproxy.pool_request.assert_has_calls(delete_lun_expected)

        # Volume has snapshots
        msg = ("cannot destroy 'Pool-0/{vol}': volume has children\nuse '-r'"
               " to destroy the following datasets:\nPool-0/{vol}@s1")
        msg = msg.format(vol='v_' + UUID_1)

        url = "http://192.168.0.2:82/api/v3/pools/Pool-0/volumes/" + UUID_1
        err = {"class": "zfslib.wrap.zfs.ZfsCmdError",
               "errno": 1000,
               "message": msg,
               "url": url}

        resp = {
            'code': 500,
            'data': None,
            'error': err}

        delete_lun_expected += [mock.call('DELETE', addr)]
        jrest.rproxy.pool_request.return_value = resp
        self.assertRaises(
            exception.VolumeIsBusy,
            jrest.delete_lun,
            'v_' + UUID_1)

        jrest.rproxy.pool_request.assert_has_calls(delete_lun_expected)

    def test_delete_lun_args(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        addr = "/volumes/v_" + UUID_1

        # Delete OK
        resp = {'data': None,
                'error': None,
                'code': 204}
        req = {'recursively_children': True,
               'recursively_dependents': True,
               'force_umount': True}

        delete_lun_expected = [mock.call('DELETE', addr, json_data=req)]
        jrest.rproxy.pool_request.return_value = resp
        self.assertIsNone(
            jrest.delete_lun('v_' + UUID_1,
                             recursively_children=True,
                             recursively_dependents=True,
                             force_umount=True))

        jrest.rproxy.pool_request.assert_has_calls(delete_lun_expected)

    def test_is_target(self):

        jrest, ctx = self.get_rest(CONFIG_OK)

        tname = CONFIG_OK['iscsi_target_prefix'] + UUID_1
        addr = '/san/iscsi/targets/{}'.format(tname)
        data = {'incoming_users_active': True,
                'name': tname,
                'allow_ip': [],
                'outgoing_user': None,
                'active': True,
                'conflicted': False,
                'deny_ip': []}

        resp = {'data': data,
                'error': None,
                'code': 200}

        is_target_expected = [mock.call('GET', addr)]
        jrest.rproxy.pool_request.return_value = resp
        self.assertTrue(jrest.is_target(tname))

        msg = "Target {} not exists.".format(tname)
        url = ("http://{addr}:{port}/api/v3/pools/Pool-0/"
               "san/iscsi/targets/{target}")
        url = url.format(addr=CONFIG_OK['san_hosts'][0],
                         port=CONFIG_OK['san_api_port'],
                         target=tname)
        err = {"class": "opene.exceptions.ItemNotFoundError",
               "message": msg,
               "url": url}

        resp = {'data': None,
                'error': err,
                'code': 404}

        is_target_expected += [mock.call('GET', addr)]
        jrest.rproxy.pool_request.return_value = resp
        self.assertEqual(False, jrest.is_target(tname))

        jrest.rproxy.pool_request.assert_has_calls(is_target_expected)

    def test_create_target(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        # Create OK
        tname = CONFIG_OK['iscsi_target_prefix'] + UUID_1
        addr = '/san/iscsi/targets'
        data = {'incoming_users_active': True,
                'name': tname,
                'allow_ip': [],
                'outgoing_user': None,
                'active': True,
                'conflicted': False,
                'deny_ip': []}

        resp = {'data': data,
                'error': None,
                'code': 201}

        req = {'name': tname,
               'active': True,
               'incoming_users_active': True}

        jrest.rproxy.pool_request.return_value = resp
        create_target_expected = [mock.call('POST', addr, json_data=req)]
        self.assertIsNone(jrest.create_target(tname))

        # Target exists
        tname = CONFIG_OK['iscsi_target_prefix'] + UUID_1
        addr = '/san/iscsi/targets'
        data = {'incoming_users_active': True,
                'name': tname,
                'allow_ip': [],
                'outgoing_user': None,
                'active': True,
                'conflicted': False,
                'deny_ip': []}

        resp = {'data': data,
                'error': None,
                'code': 201}

        url = ("http://{addr}:{port}/api/v3/pools/Pool-0/"
               "san/iscsi/targets")
        url = url.format(addr=CONFIG_OK['san_hosts'][0],
                         port=CONFIG_OK['san_api_port'])
        msg = "Target with name {} is already present on Pool-0.".format(tname)

        err = {"class": "opene.san.target.base.iscsi.TargetNameConflictError",
               "message": msg,
               "url": url}

        resp = {'data': None,
                'error': err,
                'code': 409}

        jrest.rproxy.pool_request.return_value = resp
        create_target_expected += [mock.call('POST', addr, json_data=req)]

        self.assertRaises(jexc.JDSSResourceExistsException,
                          jrest.create_target, tname)

        # Unknown error
        tname = CONFIG_OK['iscsi_target_prefix'] + UUID_1
        addr = "/san/iscsi/targets"

        resp = {'data': data,
                'error': None,
                'code': 500}

        url = ("http://{addr}:{port}/api/v3/pools/Pool-0/"
               "san/iscsi/targets")
        url = url.format(addr=CONFIG_OK['san_hosts'][0],
                         port=CONFIG_OK['san_api_port'])

        msg = "Target with name {} faced some fatal failure.".format(tname)

        err = {"class": "some test error",
               "message": msg,
               "url": url,
               "errno": 123}

        resp = {'data': None,
                'error': err,
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        create_target_expected += [mock.call('POST', addr, json_data=req)]

        self.assertRaises(jexc.JDSSException,
                          jrest.create_target, tname)

        jrest.rproxy.pool_request.assert_has_calls(create_target_expected)

    def test_delete_target(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        # Delete OK
        tname = CONFIG_OK['iscsi_target_prefix'] + UUID_1
        addr = '/san/iscsi/targets/{}'.format(tname)

        resp = {'data': None,
                'error': None,
                'code': 204}

        jrest.rproxy.pool_request.return_value = resp
        delete_target_expected = [mock.call('DELETE', addr)]
        self.assertIsNone(jrest.delete_target(tname))

        # Delete no such target

        url = ("http://{addr}:{port}/api/v3/pools/Pool-0/"
               "san/iscsi/targets")
        url = url.format(addr=CONFIG_OK['san_hosts'][0],
                         port=CONFIG_OK['san_api_port'])
        err = {"class": "opene.exceptions.ItemNotFoundError",
               "message": "Target {} not exists.".format(tname),
               "url": url}

        resp = {'data': None,
                'error': err,
                'code': 404}

        jrest.rproxy.pool_request.return_value = resp
        delete_target_expected += [mock.call('DELETE', addr)]

        self.assertRaises(jexc.JDSSResourceNotFoundException,
                          jrest.delete_target, tname)
        # Delete unknown error
        err = {"class": "some test error",
               "message": "test error message",
               "url": url,
               "errno": 123}

        resp = {'data': None,
                'error': err,
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        delete_target_expected += [mock.call('DELETE', addr)]

        self.assertRaises(jexc.JDSSException,
                          jrest.delete_target, tname)

        jrest.rproxy.pool_request.assert_has_calls(delete_target_expected)

    def test_create_target_user(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        # Modify OK
        tname = CONFIG_OK['iscsi_target_prefix'] + UUID_1
        addr = '/san/iscsi/targets/{}/incoming-users'.format(tname)

        chap_cred = {"name": "chapuser",
                     "password": "123456789012"}
        resp = {'data': None,
                'error': None,
                'code': 201}

        jrest.rproxy.pool_request.return_value = resp
        expected = [mock.call('POST', addr, json_data=chap_cred)]
        self.assertIsNone(jrest.create_target_user(tname, chap_cred))

        # No such target

        url = ("http://{addr}:{port}/api/v3/pools/Pool-0/"
               "san/iscsi/targets")
        url = url.format(addr=CONFIG_OK['san_hosts'][0],
                         port=CONFIG_OK['san_api_port'])
        err = {"class": "opene.exceptions.ItemNotFoundError",
               "message": "Target {} not exists.".format(tname),
               "url": url}

        resp = {'data': None,
                'error': err,
                'code': 404}

        jrest.rproxy.pool_request.return_value = resp
        expected += [mock.call('POST', addr, json_data=chap_cred)]

        self.assertRaises(jexc.JDSSResourceNotFoundException,
                          jrest.create_target_user, tname, chap_cred)
        # Unknown error
        err = {"class": "some test error",
               "message": "test error message",
               "url": url,
               "errno": 123}

        resp = {'data': None,
                'error': err,
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        expected += [mock.call('POST', addr, json_data=chap_cred)]

        self.assertRaises(jexc.JDSSException,
                          jrest.create_target_user, tname, chap_cred)

        jrest.rproxy.pool_request.assert_has_calls(expected)

    def test_get_target_user(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        # Get OK
        tname = CONFIG_OK['iscsi_target_prefix'] + UUID_1
        addr = '/san/iscsi/targets/{}/incoming-users'.format(tname)

        chap_users = {"name": "chapuser"}

        resp = {'data': chap_users,
                'error': None,
                'code': 200}

        jrest.rproxy.pool_request.return_value = resp
        get_target_user_expected = [mock.call('GET', addr)]
        self.assertEqual(chap_users, jrest.get_target_user(tname))

        # No such target

        url = ("http://{addr}:{port}/api/v3/pools/Pool-0/"
               "san/iscsi/targets")
        url = url.format(addr=CONFIG_OK['san_hosts'][0],
                         port=CONFIG_OK['san_api_port'])
        err = {"class": "opene.exceptions.ItemNotFoundError",
               "message": "Target {} not exists.".format(tname),
               "url": url}

        resp = {'data': None,
                'error': err,
                'code': 404}

        jrest.rproxy.pool_request.return_value = resp
        get_target_user_expected += [mock.call('GET', addr)]

        self.assertRaises(jexc.JDSSResourceNotFoundException,
                          jrest.get_target_user, tname)
        # Unknown error
        err = {"class": "some test error",
               "message": "test error message",
               "url": url,
               "errno": 123}

        resp = {'data': None,
                'error': err,
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        get_target_user_expected += [mock.call('GET', addr)]

        self.assertRaises(jexc.JDSSException,
                          jrest.get_target_user, tname)

        jrest.rproxy.pool_request.assert_has_calls(get_target_user_expected)

    def test_delete_target_user(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        # Delete OK
        tname = CONFIG_OK['iscsi_target_prefix'] + UUID_1
        user = "chapuser"
        addr = '/san/iscsi/targets/{}/incoming-users/chapuser'.format(tname)

        resp = {'data': None,
                'error': None,
                'code': 204}

        jrest.rproxy.pool_request.return_value = resp
        delete_target_user_expected = [mock.call('DELETE', addr)]
        self.assertIsNone(jrest.delete_target_user(tname, user))

        # No such user

        url = ("http://{addr}:{port}/api/v3/pools/Pool-0/"
               "san/iscsi/targets/{tname}/incoming-user/{chapuser}")
        url = url.format(addr=CONFIG_OK['san_hosts'][0],
                         port=CONFIG_OK['san_api_port'],
                         tname=tname,
                         chapuser=user)
        err = {"class": "opene.exceptions.ItemNotFoundError",
               "message": "User {} not exists.".format(user),
               "url": url}

        resp = {'data': None,
                'error': err,
                'code': 404}

        jrest.rproxy.pool_request.return_value = resp
        delete_target_user_expected += [mock.call('DELETE', addr)]

        self.assertRaises(jexc.JDSSResourceNotFoundException,
                          jrest.delete_target_user, tname, user)
        # Unknown error
        err = {"class": "some test error",
               "message": "test error message",
               "url": url,
               "errno": 123}

        resp = {'data': None,
                'error': err,
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        delete_target_user_expected += [mock.call('DELETE', addr)]

        self.assertRaises(jexc.JDSSException,
                          jrest.delete_target_user, tname, user)

        jrest.rproxy.pool_request.assert_has_calls(delete_target_user_expected)

    def test_is_target_lun(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        # lun present
        tname = CONFIG_OK['iscsi_target_prefix'] + UUID_1
        vname = jcom.vname(UUID_1)
        addr = '/san/iscsi/targets/{target}/luns/{lun}'.format(
            target=tname, lun=vname)
        data = {
            "block_size": 512,
            "device_handler": "vdisk_fileio",
            "lun": 0,
            "mode": "wt",
            "name": vname,
            "prod_id": "Storage",
            "scsi_id": "99e2c883331edf87"}
        resp = {'data': data,
                'error': None,
                'code': 200}

        jrest.rproxy.pool_request.return_value = resp
        is_target_lun_expected = [mock.call('GET', addr)]
        self.assertTrue(jrest.is_target_lun(tname, vname))

        url = "http://{ip}:{port}/api/v3/pools/Pool-0{addr}"
        url = url.format(ip=CONFIG_OK['san_hosts'][0],
                         port=CONFIG_OK['san_api_port'],
                         tname=tname,
                         addr=addr)
        msg = "volume name {lun} is not attached to target {target}"
        msg = msg.format(lun=vname, target=tname)
        err = {"class": "opene.exceptions.ItemNotFoundError",
               "message": msg,
               "url": url}

        resp = {'data': None,
                'error': err,
                'code': 404}

        jrest.rproxy.pool_request.return_value = resp
        is_target_lun_expected += [mock.call('GET', addr)]

        self.assertEqual(False, jrest.is_target_lun(tname, vname))

        err = {"class": "some test error",
               "message": "test error message",
               "url": url,
               "errno": 123}

        resp = {'data': None,
                'error': err,
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        is_target_lun_expected += [mock.call('GET', addr)]

        self.assertRaises(jexc.JDSSException,
                          jrest.is_target_lun, tname, vname)

        jrest.rproxy.pool_request.assert_has_calls(is_target_lun_expected)

    def test_attach_target_vol(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        # attach ok
        tname = CONFIG_OK['iscsi_target_prefix'] + UUID_1
        vname = jcom.vname(UUID_1)

        addr = '/san/iscsi/targets/{}/luns'.format(tname)
        jbody = {"name": vname, "lun": 0}

        data = {"block_size": 512,
                "device_handler": "vdisk_fileio",
                "lun": 0,
                "mode": "wt",
                "name": vname,
                "prod_id": "Storage",
                "scsi_id": "99e2c883331edf87"}

        resp = {'data': data,
                'error': None,
                'code': 201}

        jrest.rproxy.pool_request.return_value = resp
        attach_target_vol_expected = [
            mock.call('POST', addr, json_data=jbody)]
        self.assertIsNone(jrest.attach_target_vol(tname, vname))

        # lun attached already
        url = 'http://85.14.118.246:11582/api/v3/pools/Pool-0/{}'.format(addr)
        msg = 'Volume /dev/Pool-0/{} is already used.'.format(vname)
        err = {"class": "opene.exceptions.ItemConflictError",
               "message": msg,
               "url": url}

        resp = {'data': None,
                'error': err,
                'code': 409}

        jrest.rproxy.pool_request.return_value = resp
        attach_target_vol_expected += [
            mock.call('POST', addr, json_data=jbody)]
        self.assertRaises(jexc.JDSSResourceExistsException,
                          jrest.attach_target_vol, tname, vname)

        # no such target
        url = 'http://85.14.118.246:11582/api/v3/pools/Pool-0/{}'.format(addr)
        msg = 'Target {} not exists.'.format(vname)
        err = {"class": "opene.exceptions.ItemNotFoundError",
               "message": msg,
               "url": url}

        resp = {'data': None,
                'error': err,
                'code': 404}

        jrest.rproxy.pool_request.return_value = resp
        attach_target_vol_expected += [
            mock.call('POST', addr, json_data=jbody)]
        self.assertRaises(jexc.JDSSResourceNotFoundException,
                          jrest.attach_target_vol, tname, vname)

        # error unknown
        url = 'http://85.14.118.246:11582/api/v3/pools/Pool-0/{}'.format(addr)
        msg = 'Target {} not exists.'.format(vname)

        err = {"class": "some test error",
               "message": "test error message",
               "url": url,
               "errno": 123}

        resp = {'data': None,
                'error': err,
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        attach_target_vol_expected += [
            mock.call('POST', addr, json_data=jbody)]
        self.assertRaises(jexc.JDSSException,
                          jrest.attach_target_vol, tname, vname)
        jrest.rproxy.pool_request.assert_has_calls(attach_target_vol_expected)

    def test_detach_target_vol(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        # detach target vol ok
        tname = CONFIG_OK['iscsi_target_prefix'] + UUID_1
        vname = jcom.vname(UUID_1)

        addr = '/san/iscsi/targets/{tar}/luns/{vol}'.format(
            tar=tname, vol=vname)

        resp = {'data': None,
                'error': None,
                'code': 204}

        jrest.rproxy.pool_request.return_value = resp
        detach_target_vol_expected = [
            mock.call('DELETE', addr)]
        self.assertIsNone(jrest.detach_target_vol(tname, vname))

        # no such target
        url = 'http://85.14.118.246:11582/api/v3/pools/Pool-0/{}'.format(addr)
        msg = 'Target {} not exists.'.format(vname)
        err = {"class": "opene.exceptions.ItemNotFoundError",
               "message": msg,
               "url": url}

        resp = {'data': None,
                'error': err,
                'code': 404}

        jrest.rproxy.pool_request.return_value = resp
        detach_target_vol_expected += [
            mock.call('DELETE', addr)]
        self.assertRaises(jexc.JDSSResourceNotFoundException,
                          jrest.detach_target_vol, tname, vname)

        # error unknown
        url = 'http://85.14.118.246:11582/api/v3/pools/Pool-0/{}'.format(addr)
        msg = 'Target {} not exists.'.format(vname)

        err = {"class": "some test error",
               "message": "test error message",
               "url": url,
               "errno": 125}

        resp = {'data': None,
                'error': err,
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        detach_target_vol_expected += [
            mock.call('DELETE', addr)]
        self.assertRaises(jexc.JDSSException,
                          jrest.detach_target_vol, tname, vname)
        jrest.rproxy.pool_request.assert_has_calls(detach_target_vol_expected)

    def test_create_snapshot(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        vname = jcom.vname(UUID_1)
        sname = jcom.sname(UUID_2)

        data = {'name': jcom.sname(UUID_2)}
        resp = {'data': data,
                'error': None,
                'code': 201}

        jrest.rproxy.pool_request.return_value = resp
        self.assertIsNone(jrest.create_snapshot(vname, sname))

    def test_create_snapshot_exception(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        vname = jcom.vname(UUID_1)
        sname = jcom.sname(UUID_2)

        addr = '/volumes/{vol}/snapshots'.format(vol=vname)
        req = {'snapshot_name': sname}

        url = ('http://192.168.0.2:82/api/v3/pools/Pool-0/volumes/{vol}/'
               'snapshots').format(vol=UUID_1)
        resp = {'data': None,
                'error': {
                    'class': "zfslib.zfsapi.resources.ZfsResourceError",
                    'errno': 1,
                    'message': ('Zfs resource: Pool-0/{vol} not found in '
                                'this collection.'.format(vol=vname)),
                    "url": url},
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        create_snapshot_expected = [
            mock.call('POST', addr, json_data=req)]

        self.assertRaises(jexc.JDSSVolumeNotFoundException,
                          jrest.create_snapshot,
                          vname,
                          sname)

        # snapshot exists
        resp = {'data': None,
                'error': {
                    'class': "zfslib.zfsapi.resources.ZfsResourceError",
                    'errno': 5,
                    'message': 'Resource Pool-0/{vol}@{snap} already exists.',
                    'url': url},
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        create_snapshot_expected += [mock.call('POST', addr, json_data=req)]
        self.assertRaises(jexc.JDSSSnapshotExistsException,
                          jrest.create_snapshot,
                          vname,
                          sname)

        # error unknown
        err = {"class": "some test error",
               "message": "test error message",
               "url": url,
               "errno": 123}

        resp = {'data': None,
                'error': err,
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        create_snapshot_expected += [mock.call('POST', addr, json_data=req)]
        self.assertRaises(jexc.JDSSException,
                          jrest.create_snapshot,
                          vname,
                          sname)
        jrest.rproxy.pool_request.assert_has_calls(create_snapshot_expected)

    def test_create_volume_from_snapshot(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        vname = jcom.vname(UUID_1)
        sname = jcom.sname(UUID_2)
        cname = jcom.vname(UUID_3)

        addr = '/volumes/{vol}/clone'.format(vol=vname)
        jbody = {
            'name': cname,
            'snapshot': sname,
            'sparse': False
        }

        data = {
            "origin": "Pool-0/{vol}@{snap}".format(vol=vname, snap=sname),
            "is_clone": True,
            "full_name": "Pool-0/{}".format(cname),
            "name": cname
        }

        resp = {'data': data,
                'error': None,
                'code': 201}

        jrest.rproxy.pool_request.return_value = resp
        create_volume_from_snapshot_expected = [
            mock.call('POST', addr, json_data=jbody)]
        self.assertIsNone(jrest.create_volume_from_snapshot(cname,
                                                            sname,
                                                            vname))

        jrest.rproxy.pool_request.assert_has_calls(
            create_volume_from_snapshot_expected)

    def test_create_volume_from_snapshot_exception(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        vname = jcom.vname(UUID_1)
        sname = jcom.sname(UUID_2)
        cname = jcom.vname(UUID_3)

        addr = '/volumes/{vol}/clone'.format(vol=vname)
        jbody = {
            'name': cname,
            'snapshot': sname,
            'sparse': False
        }

        # volume DNE
        url = ('http://192.168.0.2:82/api/v3/pools/Pool-0/volumes/{vol}/'
               'clone').format(vol=UUID_1)
        resp = {'data': None,
                'error': {
                    'class': "zfslib.zfsapi.resources.ZfsResourceError",
                    'errno': 1,
                    'message': ('Zfs resource: Pool-0/{vol} not found in '
                                'this collection.'.format(vol=vname)),
                    "url": url},
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        create_volume_from_snapshot_expected = [
            mock.call('POST', addr, json_data=jbody)]

        self.assertRaises(jexc.JDSSResourceNotFoundException,
                          jrest.create_volume_from_snapshot,
                          cname,
                          sname,
                          vname)

        # clone exists
        resp = {'data': None,
                'error': {
                    "class": "zfslib.wrap.zfs.ZfsCmdError",
                    "errno": 100,
                    "message": ("cannot create 'Pool-0/{}': "
                                "dataset already exists").format(vname),
                    'url': url},
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        create_volume_from_snapshot_expected += [
            mock.call('POST', addr, json_data=jbody)]
        self.assertRaises(jexc.JDSSResourceExistsException,
                          jrest.create_volume_from_snapshot,
                          cname,
                          sname,
                          vname)

        # error unknown
        err = {"class": "some test error",
               "message": "test error message",
               "url": url,
               "errno": 123}

        resp = {'data': None,
                'error': err,
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        create_volume_from_snapshot_expected += [
            mock.call('POST', addr, json_data=jbody)]
        self.assertRaises(jexc.JDSSException,
                          jrest.create_volume_from_snapshot,
                          cname,
                          sname,
                          vname)
        jrest.rproxy.pool_request.assert_has_calls(
            create_volume_from_snapshot_expected)

    def test_rollback_volume_to_snapshot(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        vname = jcom.vname(UUID_1)
        sname = jcom.sname(UUID_2)

        req = ('/volumes/{vol}/snapshots/'
               '{snap}/rollback').format(vol=vname, snap=sname)

        resp = {'data': None,
                'error': None,
                'code': 200}

        jrest.rproxy.pool_request.return_value = resp
        rollback_volume_to_snapshot_expected = [
            mock.call('POST', req)]
        self.assertIsNone(jrest.rollback_volume_to_snapshot(vname, sname))

        jrest.rproxy.pool_request.assert_has_calls(
            rollback_volume_to_snapshot_expected)

    def test_rollback_volume_to_snapshot_exception(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        vname = jcom.vname(UUID_1)
        sname = jcom.sname(UUID_2)

        req = ('/volumes/{vol}/snapshots/'
               '{snap}/rollback').format(vol=vname,
                                         snap=sname)

        # volume DNE
        msg = ('Zfs resource: Pool-0/{vname}'
               ' not found in this collection.').format(vname=vname)

        url = ('http://192.168.0.2:82/api/v3/pools/Pool-0/volumes/{vol}/'
               'snapshots/{snap}/rollback').format(vol=vname, snap=sname)
        err = {"class": "zfslib.zfsapi.resources.ZfsResourceError",
               "message": msg,
               "url": url,
               "errno": 123}

        resp = {'data': None,
                'error': err,
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        rollback_volume_to_snapshot_expected = [
            mock.call('POST', req)]
        self.assertRaises(jexc.JDSSException,
                          jrest.rollback_volume_to_snapshot,
                          vname,
                          sname)
        jrest.rproxy.pool_request.assert_has_calls(
            rollback_volume_to_snapshot_expected)

        # error unknown
        err = {"class": "some test error",
               "message": "test error message",
               "url": url,
               "errno": 123}

        resp = {'data': None,
                'error': err,
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        rollback_volume_to_snapshot_expected += [
            mock.call('POST', req)]
        self.assertRaises(jexc.JDSSException,
                          jrest.rollback_volume_to_snapshot,
                          vname,
                          sname)
        jrest.rproxy.pool_request.assert_has_calls(
            rollback_volume_to_snapshot_expected)

    def test_delete_snapshot(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        vname = jcom.vname(UUID_1)
        sname = jcom.sname(UUID_2)

        addr = '/volumes/{vol}/snapshots/{snap}'.format(vol=vname, snap=sname)

        jbody = {
            'recursively_children': True,
            'recursively_dependents': True,
            'force_umount': True
        }

        resp = {'data': None,
                'error': None,
                'code': 204}

        jrest.rproxy.pool_request.return_value = resp
        delete_snapshot_expected = [mock.call('DELETE', addr)]
        self.assertIsNone(jrest.delete_snapshot(vname, sname))

        delete_snapshot_expected += [
            mock.call('DELETE', addr, json_data=jbody)]
        self.assertIsNone(jrest.delete_snapshot(vname,
                                                sname,
                                                recursively_children=True,
                                                recursively_dependents=True,
                                                force_umount=True))

        jrest.rproxy.pool_request.assert_has_calls(delete_snapshot_expected)

    def test_delete_snapshot_exception(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        vname = jcom.vname(UUID_1)
        sname = jcom.sname(UUID_2)
        cname = jcom.sname(UUID_3)

        addr = '/volumes/{vol}/snapshots/{snap}'.format(vol=vname, snap=sname)

        # snapshot busy
        url = ('http://192.168.0.2:82/api/v3/pools/Pool-0/volumes/{vol}/'
               'snapshots/{snap}').format(vol=vname, snap=sname)
        msg = ('cannot destroy "Pool-0/{vol}@{snap}": snapshot has dependent '
               'clones use "-R" to destroy the following datasets: '
               'Pool-0/{clone}').format(vol=vname, snap=sname, clone=cname)
        err = {'class': 'zfslib.wrap.zfs.ZfsCmdError',
               'message': msg,
               'url': url,
               'errno': 1000}

        resp = {'data': None,
                'error': err,
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        delete_snapshot_expected = [
            mock.call('DELETE', addr)]

        self.assertRaises(jexc.JDSSSnapshotIsBusyException,
                          jrest.delete_snapshot,
                          vname,
                          sname)

        # error unknown
        err = {"class": "some test error",
               "message": "test error message",
               "url": url,
               "errno": 123}

        resp = {'data': None,
                'error': err,
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        delete_snapshot_expected += [mock.call('DELETE', addr)]
        self.assertRaises(jexc.JDSSException,
                          jrest.delete_snapshot, vname, sname)

        jrest.rproxy.pool_request.assert_has_calls(delete_snapshot_expected)

    def test_get_snapshots(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        vname = jcom.vname(UUID_1)

        addr = '/volumes/{vol}/snapshots'.format(vol=vname)

        data = {"results": 2,
                "entries": {"referenced": "65536",
                            "name": jcom.sname(UUID_2),
                            "defer_destroy": "off",
                            "userrefs": "0",
                            "primarycache": "all",
                            "type": "snapshot",
                            "creation": "2015-5-27 16:8:35",
                            "refcompressratio": "1.00x",
                            "compressratio": "1.00x",
                            "written": "65536",
                            "used": "0",
                            "clones": "",
                            "mlslabel": "none",
                            "secondarycache": "all"}}

        resp = {'data': data,
                'error': None,
                'code': 200}

        jrest.rproxy.pool_request.return_value = resp
        get_snapshots_expected = [mock.call('GET', addr)]
        self.assertEqual(data['entries'], jrest.get_snapshots(vname))

        jrest.rproxy.pool_request.assert_has_calls(get_snapshots_expected)

    def test_get_snapshots_exception(self):

        jrest, ctx = self.get_rest(CONFIG_OK)
        vname = jcom.vname(UUID_1)

        addr = '/volumes/{vol}/snapshots'.format(vol=vname)

        url = ('http://192.168.0.2:82/api/v3/pools/Pool-0/volumes/{vol}/'
               'snapshots').format(vol=vname)

        err = {"class": "zfslib.zfsapi.resources.ZfsResourceError",
               "message": ('Zfs resource: Pool-0/{vol} not found in '
                           'this collection.').format(vol=vname),
               "url": url,
               "errno": 1}

        resp = {'data': None,
                'error': err,
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        get_snapshots_expected = [mock.call('GET', addr)]
        self.assertRaises(jexc.JDSSResourceNotFoundException,
                          jrest.get_snapshots,
                          vname)

        # error unknown
        err = {"class": "some test error",
               "message": "test error message",
               "url": url,
               "errno": 123}

        resp = {'data': None,
                'error': err,
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        get_snapshots_expected += [
            mock.call('GET', addr)]
        self.assertRaises(jexc.JDSSException, jrest.get_snapshots, vname)

        jrest.rproxy.pool_request.assert_has_calls(get_snapshots_expected)

    def test_get_pool_stats(self):

        jrest, ctx = self.get_rest(CONFIG_OK)

        addr = ''

        data = {"available": "950040707072",
                "status": 26,
                "name": "Pool-0",
                "scan": None,
                "encryption": {"enabled": False},
                "iostats": {
                    "read": "0",
                    "write": "0",
                    "chksum": "0"},
                "vdevs": [{"name": "wwn-0x5000cca3a8cddb2f",
                           "iostats": {"read": "0",
                                       "write": "0",
                                       "chksum": "0"},
                           "disks": [{"origin": "local",
                                      "led": "off",
                                      "name": "sdc",
                                      "iostats": {"read": "0",
                                                  "write": "0",
                                                  "chksum": "0"},
                                      "health": "ONLINE",
                                      "sn": "JPW9K0N20ZGXWE",
                                      "path": None,
                                      "model": "Hitachi HUA72201",
                                      "id": "wwn-0x5000cca3a8cddb2f",
                                      "size": 1000204886016}],
                           "health": "ONLINE",
                           "vdev_replacings": [],
                           "vdev_spares": [],
                           "type": ""}],
                "health": "ONLINE",
                "operation": "none",
                "id": "12413634663904564349",
                "size": "996432412672"}

        resp = {'data': data,
                'error': None,
                'code': 200}

        jrest.rproxy.pool_request.return_value = resp
        get_pool_stats_expected = [mock.call('GET', addr)]
        self.assertEqual(data, jrest.get_pool_stats())

        jrest.rproxy.pool_request.assert_has_calls(get_pool_stats_expected)

    def test_get_pool_stats_exception(self):

        jrest, ctx = self.get_rest(CONFIG_OK)

        addr = ''

        url = 'http://192.168.0.2:82/api/v3/pools/Pool-0/'

        err = {'class': 'zfslib.zfsapi.zpool.ZpoolError',
               'message': "Given zpool 'Pool-0' doesn't exists.",
               "url": url,
               "errno": 1}

        resp = {'data': None,
                'error': err,
                'code': 500}

        jrest.rproxy.pool_request.return_value = resp
        get_pool_stats_expected = [mock.call('GET', addr)]
        self.assertRaises(jexc.JDSSException, jrest.get_pool_stats)

        jrest.rproxy.pool_request.assert_has_calls(get_pool_stats_expected)



# Copyright 2016 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The volume metadata V3 api."""

from http import HTTPStatus

from oslo_serialization import jsonutils
from oslo_utils.secretutils import md5
import webob

from cinder.api import microversions as mv
from cinder.api.openstack import wsgi
from cinder.api.schemas import volume_metadata as metadata
from cinder.api.v2 import volume_metadata as volume_meta_v2
from cinder.api import validation


class Controller(volume_meta_v2.Controller):
    """The volume metadata API controller for the OpenStack API."""
    def _validate_etag(self, req, volume_id):
        if not req.if_match:
            return True
        context = req.environ['cinder.context']
        metadata = self._get_metadata(context, volume_id)
        data = jsonutils.dumps({"metadata": metadata})
        data = data.encode('utf-8')
        checksum = md5(data, usedforsecurity=False).hexdigest()
        return checksum in req.if_match.etags

    @wsgi.extends
    def index(self, req, volume_id):
        req_version = req.api_version_request
        metadata = super(Controller, self).index(req, volume_id)
        if req_version.matches(mv.ETAGS):
            data = jsonutils.dumps(metadata)
            data = data.encode('utf-8')
            resp = webob.Response()
            resp.headers['Etag'] = md5(data, usedforsecurity=False).hexdigest()
            resp.body = data
            return resp
        return metadata

    @wsgi.extends
    @validation.schema(metadata.update)
    def update(self, req, volume_id, id, body):
        req_version = req.api_version_request
        if req_version.matches(mv.ETAGS):
            if not self._validate_etag(req, volume_id):
                return webob.Response(
                    status_int=HTTPStatus.PRECONDITION_FAILED)
        return super(Controller, self).update(req, volume_id,
                                              id, body=body)

    @wsgi.extends
    @validation.schema(metadata.create)
    def update_all(self, req, volume_id, body):
        req_version = req.api_version_request
        if req_version.matches(mv.ETAGS):
            if not self._validate_etag(req, volume_id):
                return webob.Response(
                    status_int=HTTPStatus.PRECONDITION_FAILED)
        return super(Controller, self).update_all(req, volume_id,
                                                  body=body)


def create_resource():
    return wsgi.Resource(Controller())



# Copyright (C) 2015 Tom Barron <tpb@dyncloud.net>
# Copyright (C) 2015 Kevin Fox <kevin@efox.cc>
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Implementation of a backup service that uses a posix filesystem as the
   backend."""

import errno
import os
import os.path
import stat

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import timeutils

from cinder.backup import chunkeddriver
from cinder import exception
from cinder import interface

LOG = logging.getLogger(__name__)

SHA_SIZE = 32768
# Multiple of SHA_SIZE, close to a characteristic OS max file system size.
BACKUP_FILE_SIZE = 61035 * 32768

posixbackup_service_opts = [
    cfg.IntOpt('backup_file_size',
               default=BACKUP_FILE_SIZE,
               help='The maximum size in bytes of the files used to hold '
                    'backups. If the volume being backed up exceeds this '
                    'size, then it will be backed up into multiple files.'
                    'backup_file_size must be a multiple of '
                    'backup_sha_block_size_bytes.'),
    cfg.IntOpt('backup_sha_block_size_bytes',
               default=SHA_SIZE,
               help='The size in bytes that changes are tracked '
                    'for incremental backups. backup_file_size has '
                    'to be multiple of backup_sha_block_size_bytes.'),
    cfg.BoolOpt('backup_enable_progress_timer',
                default=True,
                help='Enable or Disable the timer to send the periodic '
                     'progress notifications to Ceilometer when backing '
                     'up the volume to the backend storage. The '
                     'default value is True to enable the timer.'),
    cfg.StrOpt('backup_posix_path',
               default='$state_path/backup',
               help='Path specifying where to store backups.'),
    cfg.StrOpt('backup_container',
               help='Custom directory to use for backups.'),
]

CONF = cfg.CONF
CONF.register_opts(posixbackup_service_opts)


@interface.backupdriver
class PosixBackupDriver(chunkeddriver.ChunkedBackupDriver):
    """Provides backup, restore and delete using a Posix file system."""

    def __init__(self, context, backup_path=None):
        chunk_size_bytes = CONF.backup_file_size
        sha_block_size_bytes = CONF.backup_sha_block_size_bytes
        backup_default_container = CONF.backup_container
        enable_progress_timer = CONF.backup_enable_progress_timer
        super().__init__(
            context,
            chunk_size_bytes,
            sha_block_size_bytes,
            backup_default_container,
            enable_progress_timer,
        )
        self.backup_path = backup_path
        if not backup_path:
            self.backup_path = CONF.backup_posix_path
        if not self.backup_path:
            raise exception.ConfigNotFound(path='backup_path')
        LOG.debug("Using backup repository: %s", self.backup_path)

    @staticmethod
    def get_driver_options():
        return posixbackup_service_opts

    def update_container_name(self, backup, container):
        if container is not None:
            return container
        id = backup['id']
        return os.path.join(id[0:2], id[2:4], id)

    def put_container(self, container):
        path = os.path.join(self.backup_path, container)
        if not os.path.exists(path):
            os.makedirs(path)
            permissions = (
                stat.S_IRUSR |
                stat.S_IWUSR |
                stat.S_IXUSR |
                stat.S_IRGRP |
                stat.S_IWGRP |
                stat.S_IXGRP)
            os.chmod(path, permissions)

    def get_container_entries(self, container, prefix):
        path = os.path.join(self.backup_path, container)
        return [i for i in os.listdir(path) if i.startswith(prefix)]

    def get_object_writer(self, container, object_name, extra_metadata=None):
        path = os.path.join(self.backup_path, container, object_name)
        f = open(path, 'wb')
        permissions = (
            stat.S_IRUSR |
            stat.S_IWUSR |
            stat.S_IRGRP |
            stat.S_IWGRP)
        os.chmod(path, permissions)
        return f

    def get_object_reader(self, container, object_name, extra_metadata=None):
        path = os.path.join(self.backup_path, container, object_name)
        return open(path, 'rb')

    def delete_object(self, container, object_name):
        # TODO(tbarron):  clean up the container path if it is empty
        path = os.path.join(self.backup_path, container, object_name)
        try:
            os.remove(path)
        except OSError as e:
            # Ignore exception if path does not exist.
            if e.errno != errno.ENOENT:
                raise

    def _generate_object_name_prefix(self, backup):
        timestamp = timeutils.utcnow().strftime("%Y%m%d%H%M%S")
        prefix = 'volume_%s_%s_backup_%s' % (backup.volume_id, timestamp,
                                             backup.id)
        LOG.debug('_generate_object_name_prefix: %s', prefix)
        return prefix

    def get_extra_metadata(self, backup, volume):
        return None
